from typing import List, Optional
from pydantic import BaseModel
from typing import List

from scale_egp.utils.model_utils import BaseModel


class Account(BaseModel):
    id: str
    name: str


class UserRole(BaseModel):
    role: str
    account: Account


class UserInfoResponse(BaseModel):
    id: str
    first_name: Optional[str]
    last_name: Optional[str]
    email: str
    accounts: List[UserRole]



from typing import Optional, Literal, Iterable, Union, List

from scale_egp.sdk.types.completions import (
    CompletionRequest,
    Completion,
    ModelParameters,
    ImageCompletionRequests,
)
from scale_egp.utils.api_utils import APIEngine


class CompletionCollection(APIEngine):
    _sub_path = "v2/completions"

    def create(
        self,
        model: Union[
            Literal[
                "gpt-4",
                "gpt-4-0613",
                "gpt-4-32k",
                "gpt-4-32k-0613",
                "gpt-4-vision-preview",
                "gpt-4o",
                "gpt-3.5-turbo",
                "gpt-3.5-turbo-0613",
                "gpt-3.5-turbo-16k",
                "gpt-3.5-turbo-16k-0613",
                "text-davinci-003",
                "text-davinci-002",
                "text-curie-001",
                "text-babbage-001",
                "text-ada-001",
                "claude-instant-1",
                "claude-instant-1.1",
                "claude-2",
                "claude-2.0",
                "llama-7b",
                "llama-2-7b",
                "llama-2-7b-chat",
                "llama-2-13b",
                "llama-2-13b-chat",
                "llama-2-70b",
                "llama-2-70b-chat",
                "falcon-7b",
                "falcon-7b-instruct",
                "falcon-40b",
                "falcon-40b-instruct",
                "mpt-7b",
                "mpt-7b-instruct",
                "flan-t5-xxl",
                "mistral-7b",
                "mistral-7b-instruct",
                "mixtral-8x7b",
                "mixtral-8x7b-instruct",
                "llm-jp-13b-instruct-full",
                "llm-jp-13b-instruct-full-dolly",
                "zephyr-7b-alpha",
                "zephyr-7b-beta",
                "codellama-7b",
                "codellama-7b-instruct",
                "codellama-13b",
                "codellama-13b-instruct",
                "codellama-34b",
                "codellama-34b-instruct",
                "codellama-70b",
                "codellama-70b-instruct",
                "gemini-pro",
                "gemini-1.5-pro-preview-0409"
            ],
            str,
        ],
        prompt: str,
        account_id: str,
        images: Optional[List[ImageCompletionRequests]] = None,
        model_parameters: Optional[ModelParameters] = None,
    ) -> Completion:
        """
        Create a new LLM Completion.

        Args:
            model: The model to use for the completion.
            prompt: The prompt to use for the completion.
            model_parameters: The parameters to use for the model.

        Returns:
            The newly created Completion.
        """
        return Completion.from_dict(
            self._post(
                sub_path=self._sub_path,
                request=CompletionRequest(
                    model=model,
                    prompt=prompt,
                    images=images,
                    model_parameters=model_parameters,
                    account_id=account_id,
                ),
            ).json()
        )

    def stream(
        self,
        model: Literal[
            "gpt-4",
            "gpt-4-0613",
            "gpt-4-32k",
            "gpt-4-32k-0613",
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k",
            "gpt-3.5-turbo-16k-0613",
            "text-davinci-003",
            "text-davinci-002",
            "text-curie-001",
            "text-babbage-001",
            "text-ada-001",
            "claude-instant-1",
            "claude-instant-1.1",
            "claude-2",
            "claude-2.0",
            "llama-7b",
            "llama-2-7b",
            "llama-2-7b-chat",
            "llama-2-13b",
            "llama-2-13b-chat",
            "llama-2-70b",
            "llama-2-70b-chat",
            "falcon-7b",
            "falcon-7b-instruct",
            "falcon-40b",
            "falcon-40b-instruct",
            "mpt-7b",
            "mpt-7b-instruct",
            "flan-t5-xxl",
            "mistral-7b",
            "mistral-7b-instruct",
            "mixtral-8x7b",
            "mixtral-8x7b-instruct",
            "llm-jp-13b-instruct-full",
            "llm-jp-13b-instruct-full-dolly",
            "zephyr-7b-alpha",
            "zephyr-7b-beta",
            "codellama-7b",
            "codellama-7b-instruct",
            "codellama-13b",
            "codellama-13b-instruct",
            "codellama-34b",
            "codellama-34b-instruct",
            "codellama-70b",
            "codellama-70b-instruct",
        ],
        prompt: str,
        account_id: str,
        images: Optional[List[ImageCompletionRequests]] = None,
        model_parameters: Optional[ModelParameters] = None,
    ) -> Iterable[Completion]:
        """
        Stream LLM Completions.

        Returns:
            The newly created Completion.
        """
        iterable_payloads = self._post_stream(
            sub_path=self._sub_path,
            request=CompletionRequest(
                model=model,
                prompt=prompt,
                images=images,
                model_parameters=model_parameters,
                stream=True,
                account_id=account_id,
            ),
        )
        if iterable_payloads:
            for response_dict in iterable_payloads:
                yield Completion.from_dict(response_dict)
        else:
            return []



import logging
import sys
from typing import List, Optional
from argh import ArghParser
from scale_egp.cli.collections import (
    GenericCommands,
    CollectionCRUDCommandsForImmutable,
    EGPClientFactory,
    FineTuningJobCommands,
    ModelDeploymentCommands,
    ModelInstanceCommands,
    ModelTemplateCommands,
    TrainingDatasetCommands,
    UserCommands,
    ModelGroupCommands,
)
from scale_egp.cli.formatter import AVAILABLE_FORMATTERS
from scale_egp.cli.parser import get_parser, dispatch
import itertools
from asyncio import iscoroutinefunction, run
from functools import wraps
import traceback

from scale_egp.cli.vpc_setup_commands import AwsVpcSetupCommands


def async_to_sync(fn):
    if iscoroutinefunction(fn):

        @wraps(fn)
        def fn2(self, *args, **kwargs):
            # Note: not passing in self, since the function will be bound to that instance
            return run(fn(*args, **kwargs))

        # bind fn2 to the object to which fn was bound
        # this is important because the class instance is used by formatter to get the
        # title of the output table
        instance = fn.__self__
        bound_method = fn2.__get__(instance, instance.__class__)
        return bound_method
    return fn


def add_commands_to_parser(parser: ArghParser, commands: GenericCommands):
    parser.add_commands(
        [
            async_to_sync(getattr(commands, key))
            for key in dir(commands)
            if not key.startswith("_") and callable(getattr(commands, key))
        ],
        group_name=commands.command_group_name,
        group_kwargs={
            "title": getattr(commands, "command_group_title", None),
            "help": getattr(commands, "__doc__", None),
        },
    )


def load_plugins(parser: ArghParser):
    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points
    plugin_commands: List[CollectionCRUDCommandsForImmutable] = []
    for entrypoint in entry_points(group="scale_egp_cli"):
        try:
            plugin_commands.append((entrypoint.load())())
        except Exception as e:
            print(f"Error loading plugin: {str(e)}")
            traceback.print_exception(e)
    return itertools.chain.from_iterable(plugin_commands)


def exec_cli(argv: Optional[List[str]] = None):
    if argv is None:
        argv = sys.argv[1:]
    client_factory = EGPClientFactory()

    parser = get_parser()
    parser.add_argument("--log-curl-commands", action="store_true", default=False)
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    parser.add_argument("-k", "--api-key", type=str, default=None, metavar="EGP_API_KEY")
    parser.add_argument(
        "-i",
        "--account-id",
        type=str,
        default=None,
        help="Optional SGP account_id to use. The default account_id will be used if not set.",
        metavar="EGP_ACCOUNT_ID",
    )
    parser.add_argument("-e", "--endpoint-url", type=str, default=None, metavar="EGP_ENDPOINT_URL")
    parser.add_argument(
        "-f", "--format", type=str, choices=[*AVAILABLE_FORMATTERS.keys()], default="rich"
    )
    add_commands_to_parser(parser, ModelGroupCommands(client_factory))
    add_commands_to_parser(parser, ModelInstanceCommands(client_factory))
    add_commands_to_parser(parser, ModelDeploymentCommands(client_factory))
    add_commands_to_parser(parser, ModelTemplateCommands(client_factory))
    add_commands_to_parser(parser, UserCommands(client_factory))
    add_commands_to_parser(parser, FineTuningJobCommands(client_factory))
    add_commands_to_parser(parser, TrainingDatasetCommands(client_factory))
    add_commands_to_parser(parser, AwsVpcSetupCommands(client_factory))
    for plugin_command_cls in load_plugins(parser):
        # Uncomment to debug plugin loading issues:
        # print(f"Loading plugin: {plugin_command_cls.__name__}")
        add_commands_to_parser(parser, plugin_command_cls(client_factory))
    args = parser.parse_args(argv)
    try:
        client_factory.set_client_kwargs(
            api_key=args.api_key,
            endpoint_url=args.endpoint_url,
            account_id=args.account_id,
            log_curl_commands=args.log_curl_commands,
        )
        if args.log_curl_commands:
            logging.basicConfig(level=logging.INFO)
        return dispatch(parser, args, argv=argv)
    except Exception as e:
        if args.verbose:
            traceback.print_exception(e)
        message = getattr(e, "message", None)
        if message is None:
            message = str(e)
        status_code = ""
        if hasattr(e, "code"):
            status_code = f" (HTTP {e.code})"
        print(f"ERROR{status_code}: {message}")
        sys.exit(1)



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field

from scale_egp.sdk.enums import TestCaseSchemaType
from scale_egp.utils.model_utils import Entity, BaseModel


class EvaluationDataset(Entity):
    """
    A data model representing an evaluation dataset.

    Attributes:
        name: The name of the evaluation dataset
        schema_type: The schema type of the evaluation dataset
        id: The ID of the evaluation dataset
        created_at: The time the evaluation dataset was created
        updated_at: The time the evaluation dataset was last updated
        account_id: The ID of the account that owns the evaluation dataset
        created_by_user_id: The ID of the user that created the evaluation dataset
        version_num: The version number of the evaluation dataset

    """

    name: str
    schema_type: TestCaseSchemaType
    id: str
    created_at: datetime
    updated_at: datetime
    account_id: str
    created_by_user_id: str
    version_num: Optional[int] = None


class EvaluationDatasetRequest(BaseModel):
    name: str
    schema_type: TestCaseSchemaType
    account_id: Optional[str] = Field(
        description="Account to create knowledge base in. If you have access to more than one "
                    "account, you must specify an account_id"
    )


class EvaluationDatasetVersionRequest(BaseModel):
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")


class EvaluationDatasetVersion(Entity):
    num: int
    evaluation_dataset_id: str
    id: str
    created_at: datetime
    account_id: str
    created_by_user_id: str



from typing import List, Optional


from scale_egp.sdk.types.question_sets import QuestionSet, QuestionSetRequest
from scale_egp.sdk.types.questions import Question
from scale_egp.utils.api_utils import (
    APIEngine,
)


class QuestionSetCollection(APIEngine):
    _sub_path = "v2/question-sets"

    def create(
        self,
        name: str,
        questions: List[Question],
        account_id: Optional[str] = None,
    ) -> QuestionSet:
        """
        Create a new Question Set.

        Args:
            name: The name of the Question Set.
            questions: The questions in this Question Set.
            account_id: The ID of the account to create this Question Set for.
        Returns:
            The newly created Evaluation Configuration.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=QuestionSetRequest(
                name=name,
                account_id=account_id or self._api_client.account_id,
                question_ids=[question.id for question in questions],
            ),
        )
        return QuestionSet.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> QuestionSet:
        """
        Get the details of a Question Set.

        Args:
            id: The ID of the Question Set.

        Returns:
            The details of the Question Set.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return QuestionSet.from_dict(response.json())

    def list(
        self,
    ) -> List[QuestionSet]:
        """
        List Question Sets.

        Returns:
            A list of Question Sets.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [QuestionSet.from_dict(question_set) for question_set in response.json()]



from __future__ import annotations
from datetime import datetime

from scale_egp.sdk.enums import EvaluationType
from scale_egp.utils.model_utils import BaseModel, Entity

class EvaluationConfig(Entity):
    """
    A data model representing an evaluation config.

    Attributes:
        id: The ID of the evaluation config
        evaluation_type: The type of evaluation. Currently, only `HUMAN` is supported.
        question_set_id: The ID of the question set to use for the evaluation.
        account_id: The account ID of the account that the evaluation config is associated with.
        created_at: The time the evaluation config was created.
        created_by_user_id: The user ID of the user who created the evaluation config.
    """

    id: str
    evaluation_type: EvaluationType
    question_set_id: str
    account_id: str
    created_at: datetime
    created_by_user_id: str


class EvaluationConfigRequest(BaseModel):
    evaluation_type: EvaluationType
    question_set_id: str
    account_id: str



from .agents import *
from .application_specs import *
from .chunks import *
from .completions import *
from .embeddings import *
from .evaluation_configs import *
from .evaluation_dataset_test_cases import *
from .evaluation_datasets import *
from .evaluation_test_case_results import *
from .evaluations import *
from .knowledge_base_uploads import *
from .knowledge_base_artifacts import *
from .knowledge_base_chunks import *
from .knowledge_bases import *
from .memory_strategy import *
from .message import *
from .model_templates import *
from .models import *
from .user_info import *
from .questions import *
from .question_sets import *



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Optional, Union, List, Literal, Dict, Any

from pydantic import Field

from scale_egp.sdk.enums import (
    UploadJobStatus,
    DataSource,
    DeduplicationStrategy,
    ChunkingStrategy,
    UploadScheduleStatus,
)
from scale_egp.sdk.types.knowledge_base_artifacts import KnowledgeBaseArtifact
from scale_egp.utils.model_utils import Entity, RootModel, BaseModel


class KnowledgeBaseUploadResponse(BaseModel):
    upload_id: str = Field(..., description="ID of the created knowledge base upload job.")


class CancelKnowledgeBaseUploadResponse(BaseModel):
    upload_id: str = Field(
        ..., description="ID of the knowledge base upload job that was cancelled."
    )
    canceled: bool = Field(..., description="Whether cancellation was successful.")


class ArtifactsStatus(BaseModel):
    """
    A data model representing the status of the artifacts in a knowledge base.

    Attributes:
        artifacts_completed: Number of artifacts uploaded successfully.
        artifacts_pending: Number of artifacts awaiting upload.
        artifacts_uploading: Number of artifacts with upload in progress.
        artifacts_failed: Number of artifacts that failed upload.
    """

    artifacts_completed: int
    artifacts_pending: int
    artifacts_uploading: int
    artifacts_failed: int


class S3DataSourceConfig(BaseModel):
    """
    A data model representing the configuration of a S3 data source.

    Attributes:
        source: The data source type. Must be 's3'.
        s3_bucket: The name of the S3 bucket where the data is stored
        s3_prefix: The prefix of the S3 bucket where the data is stored
        aws_region: The AWS region where the S3 bucket is located
        aws_account_id: The AWS account ID that owns the S3 bucket
    """

    source: Literal[DataSource.S3] = DataSource.S3.value
    s3_bucket: str
    aws_region: str
    aws_account_id: str
    s3_prefix: Optional[str] = None


class SharePointDataSourceConfig(BaseModel):
    """
    A data model representing the configuration of a SharePoint data source.

    Attributes:
        source: The data source type. Must be 'sharepoint'.
        client_id: The client ID associated with this SharePoint site
        tenant_id: The tenant ID that the SharePoint site is within
        site_id: The site ID for this SharePoint site
        folder_path: The nested folder path to read files from the root of the site
        recursive: Whether to recurse through the folder contents
    """

    source: Literal[DataSource.SHAREPOINT] = DataSource.SHAREPOINT.value
    client_id: str
    tenant_id: str
    site_id: str
    folder_path: Optional[str] = ""
    recursive: Optional[bool] = True


class GoogleDriveDataSourceConfig(BaseModel):
    """
    A data model representing the configuration of a Google Drive data source.

    Attributes:
        source: The data source type. Must be 'google_drive'.
        drive_id: The ID of the Google Drive to retrieve contents from
    """

    source: Literal[DataSource.GOOGLE_DRIVE] = DataSource.GOOGLE_DRIVE.value
    drive_id: str


class LocalChunksSourceConfig(BaseModel):
    """
    A data model representing the configuration of a local chunks data source.

    Attributes:
        source: The data source type. Must be 'local_chunks'.
        artifact_name: The file name assigned to the artifact, containing a file extension.
            Adding an extension is mandatory, to allow detecting file types for text extraction.
        artifact_uri: A unique identifier for an artifact within the knowledge base, such as full
            path in a directory or file system.
        deduplication_strategy: Action to take if an artifact with the same name already exists
            in the knowledge base. Can be either Overwrite (default) or Fail.
    """

    source: Literal[DataSource.LOCAL_CHUNKS] = DataSource.LOCAL_CHUNKS.value
    artifact_name: str
    artifact_uri: str
    deduplication_strategy: Optional[DeduplicationStrategy] = DeduplicationStrategy.OVERWRITE


class GoogleDriveDataSourceAuthConfig(BaseModel):
    """
    A data model representing the configuration of a Google Drive service account.

    Attributes:
        source: The data source type. Must be 'GoogleDrive'.
        client_email: The service account's client email
        private_key: The service account's private_key
        token_uri: The service account's token_uri
        client_id: The service account's client_id
    """

    source: Literal[DataSource.GOOGLE_DRIVE] = DataSource.GOOGLE_DRIVE.value
    client_email: str
    private_key: str
    token_uri: str
    client_id: str


class SharePointDataSourceAuthConfig(BaseModel):
    """
    A data model representing the configuration of a SharePoint data source.

    Attributes:
        source: The data source type. Must be 'sharepoint'.
        client_secret: The secret for the app registration associated with this SharePoint site
    """

    source: Literal[DataSource.SHAREPOINT] = DataSource.SHAREPOINT.value
    client_secret: str


class S3DataSourceAuthConfig(BaseModel):
    """
    A data model representing the configuration of a S3 data source.

    Attributes:
        source: The data source type. Must be 'S3'.
        s3_role: Name of the role that a client will be initialized via AssumeRole of AWS sts
        external_id: External ID defined by the customer for the IAM role
    """

    source: Literal[DataSource.S3] = DataSource.S3.value
    s3_role: Optional[str]
    external_id: Optional[str]

    class Config(BaseModel.Config):
        title = "S3 DataSource Auth Config"


class AzureBlobStorageDataSourceConfig(BaseModel):
    """
    A data model representing the configuration of an Azure Blob Storage data source.

    Attributes:
        source: The data source type. Must be 'azure_blob_storage'.
        container_url: The SAS URL for the Azure Blob Storage container (a.k.a. Blob SAS URL)
    """

    source: Literal[DataSource.AZURE_BLOB_STORAGE] = DataSource.AZURE_BLOB_STORAGE.value
    container_url: str


class AzureBlobStorageDataSourceAuthConfig(BaseModel):
    """
    A data model representing the configuration of an Azure Blob Storage data source.

    Attributes:
        source: The data source type. Must be 'azure_blob_storage'.
        blob_sas_token: The SAS token for the Azure Blob Storage container
    """

    source: Literal[DataSource.AZURE_BLOB_STORAGE] = DataSource.AZURE_BLOB_STORAGE.value
    blob_sas_token: str


class DataSourceAuthConfig(RootModel):
    __root__: Union[
        GoogleDriveDataSourceAuthConfig,
        SharePointDataSourceAuthConfig,
        S3DataSourceAuthConfig,
        AzureBlobStorageDataSourceAuthConfig,
    ] = Field(..., discriminator="source")


class CharacterChunkingStrategyConfig(BaseModel):
    """
    A data model representing the configuration of a character chunking strategy.

    Attributes:
        strategy: The chunking strategy type. Must be 'character'.
        separator: Character designating breaks in input data. Text data will first be split
            into sections by this separator, then each section will be split into chunks
            of size `chunk_size`.
        chunk_size: Maximum number of characters in each chunk. If not specified, a chunk size
            of 1000 will be used.
        chunk_overlap: Number of characters to overlap between chunks. If not specified, an overlap
            of 200 will be used. For example if the chunk size is 3 and the overlap size
            is 1, and the text to chunk is 'abcde', the chunks will be 'abc', 'cde'.
    """

    strategy: Literal[ChunkingStrategy.CHARACTER] = ChunkingStrategy.CHARACTER.value
    separator: Optional[str] = "\n\n"
    chunk_size: Optional[int] = 1000
    chunk_overlap: Optional[int] = 200


class TokenChunkingStrategyConfig(BaseModel):
    """
    A data model representing the configuration of a token chunking strategy.

    Attributes:
        strategy: The chunking strategy type. Must be 'token'.
        separator: Character designating breaks in input data. Text data will first be split
            into sections by this separator, then each section will be split into chunks
            of size `chunk_size`.
        target_chunk_size: Target number of tokens in each chunk. If not specified, a target chunk
            size of 200 will be used.
        max_chunk_size: Maximum number of tokens in each chunk. If not specified, a maximum chunk
            size of 200 will be used.
        chunk_overlap: Number of tokens to overlap between chunks. If not specified, an overlap
            of 0 will be used. Note this is only followed approximately.
    """

    strategy: Literal[ChunkingStrategy.TOKEN] = ChunkingStrategy.TOKEN.value
    separator: Optional[str] = "\n\n"
    target_chunk_size: Optional[int] = 200
    max_chunk_size: Optional[int] = 600
    chunk_overlap: Optional[int] = 0


class ChunkToUpload(BaseModel):
    """
    A data model representing a local chunk.

    Attributes:
        text: The text associated with the chunk
        chunk_position: The position of the chunk in the artifact
        metadata: Any additional key value pairs of information stored with the chunk
    """

    text: str
    chunk_position: int
    metadata: Dict[str, Any] = Field(default_factory=dict)


class DataSourceConfig(RootModel):
    """
    A type alias for a Union of all data source types.

    Attributes:
        __root__: Instead of directly using this class, please use the appropriate data source type
            for your use case.
    """

    __root__: Union[
        S3DataSourceConfig,
        SharePointDataSourceConfig,
        GoogleDriveDataSourceConfig,
        AzureBlobStorageDataSourceConfig,
        LocalChunksSourceConfig,
    ] = Field(
        ...,
        discriminator="source",
    )


class RemoteDataSourceConfig(RootModel):
    """
    A type alias for a Union of all remote data source types.

    Attributes:
        __root__: Instead of directly using this class, please use the appropriate data source type
            for your use case.
    """

    __root__: Union[
        S3DataSourceConfig,
        SharePointDataSourceConfig,
        GoogleDriveDataSourceConfig,
        AzureBlobStorageDataSourceConfig,
    ] = Field(
        ...,
        discriminator="source",
    )


class ChunkingStrategyConfig(RootModel):
    """
    A type alias for a Union of all chunking strategy types.

    Attributes:
        __root__: Instead of directly using this class, please use the appropriate chunking strategy
            type for your use case.
    """

    __root__: Union[
        CharacterChunkingStrategyConfig,
        TokenChunkingStrategyConfig,
    ] = Field(
        ...,
        discriminator="strategy",
    )


class KnowledgeBaseUpload(Entity):
    """
    A data model representing a knowledge base upload.

    Attributes:
        upload_id: Unique ID of the upload job
        data_source_config: Configuration for downloading data from source
        chunking_strategy_config: Configuration for chunking the text content of each artifact
        created_at: The timestamp at which the upload job started
        updated_at: The timestamp at which the upload job was last updated
        status: Sync status
        status_reason: Reason for the upload job's status
        artifacts_status: Number of artifacts pending, completed, and failed
        artifacts: List of info for each artifacts
    """

    upload_id: str
    data_source_config: DataSourceConfig
    chunking_strategy_config: Optional[ChunkingStrategyConfig]
    created_at: str
    updated_at: str
    status: UploadJobStatus
    status_reason: Optional[str] = None
    artifacts_status: Optional[ArtifactsStatus]
    artifacts: Optional[List[KnowledgeBaseArtifact]]


class KnowledgeBaseRemoteUploadRequest(BaseModel):
    upload_type: Literal["remote"] = "remote"
    data_source_config: RemoteDataSourceConfig
    data_source_auth_config: Optional[DataSourceAuthConfig] = Field(
        None,
        description="Configuration for the data source which describes how to "
        "authenticate to the data source.",
    )
    chunking_strategy_config: ChunkingStrategyConfig = Field(
        None,
        description="Configuration for the chunking strategy which describes how to chunk the "
        "data.",
    )


class KnowledgeBaseDataSourceUploadRequest(BaseModel):
    upload_type: Literal["data_source"] = "data_source"
    data_source_id: str = Field(description="The ID of the data source to upload from.")
    chunking_strategy_config: ChunkingStrategyConfig = Field(
        None,
        description="Configuration for the chunking strategy which describes how to chunk the "
        "data.",
    )


class ListKnowledgeBaseUploadsResponse(BaseModel):
    uploads: List[KnowledgeBaseUpload] = Field(..., description="List of knowledge base uploads.")


class KnowledgeBaseLocalChunkUploadRequest(BaseModel):
    upload_type: Literal["local_chunks"] = "local_chunks"
    data_source_config: LocalChunksSourceConfig = Field(
        ...,
        description="Configuration for the data source which describes where to find the data.",
    )
    chunks: List[ChunkToUpload] = Field(..., description="List of chunks.")


class KnowledgeBaseUploadRequest(RootModel):
    __root__: Union[
        KnowledgeBaseRemoteUploadRequest,
        KnowledgeBaseLocalChunkUploadRequest,
        KnowledgeBaseDataSourceUploadRequest,
    ] = Field(
        ...,
        discriminator="upload_type",
    )


class KnowledgeBaseDataSourceRequest(BaseModel):
    name: str = Field(..., description="A unique name for the data source.")
    description: Optional[str] = Field(None, description="A description of the data source.")
    data_source_config: RemoteDataSourceConfig
    data_source_auth_config: Optional[DataSourceAuthConfig] = Field(
        None,
        description="Configuration for the data source which describes how to "
        "authenticate to the data source.",
    )
    account_id: str = Field(..., description="The ID of the account that owns the data source.")


class KnowledgeBaseDataSource(Entity):
    id: str
    name: str
    description: Optional[str]
    data_source_config: RemoteDataSourceConfig
    created_at: datetime
    updated_at: datetime
    created_by_user_id: str
    account_id: str


class KnowledgeBaseUploadScheduleRequest(BaseModel):
    knowledge_base_data_source_id: str = Field(
        ..., description="The ID of the knowledge base data source to upload from."
    )
    chunking_strategy_config: ChunkingStrategyConfig = Field(
        None,
        description="Configuration for the chunking strategy which describes how to chunk the "
        "data.",
    )
    interval: float = Field(
        ..., description="The interval at which to run uploads from the data source."
    )
    next_run_at: Optional[str] = Field(
        None, description="The time at which the next upload will run."
    )
    account_id: str = Field(..., description="The ID of the account that owns the upload schedule.")


class KnowledgeBaseUploadSchedulePauseRequest(BaseModel):
    next_run_at: None = None


class KnowledgeBaseUploadSchedule(Entity):
    id: str
    knowledge_base_id: str
    knowledge_base_data_source_id: str
    chunking_strategy_config: ChunkingStrategyConfig
    status: UploadScheduleStatus
    status_reason: Optional[str]
    interval: timedelta
    next_run_at: Optional[datetime]
    created_at: datetime
    updated_at: datetime
    created_by_user_id: str
    account_id: str



from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Field


class ToolRequest(BaseModel):
    """
    A request to run a tool.

    Attributes:
        name: Name of the tool that the AI wants the client to use.
        arguments: Arguments to pass to the tool. The format must be a JSON Schema-compliant
            object serialized into a string.
    """
    name: str = Field(...)
    arguments: str = Field(...)


class ActionContext(BaseModel):
    """
    The context of the action that the agent is taking.

    Attributes:
        content: The content of the final output of the agent when it no longer needs any tools.
        tool_request: The tool request if the agent needs more information.
    """
    content: Optional[str] = Field(default=None)
    tool_request: Optional[ToolRequest] = Field(
        description="The tool request if the agent needs more information.",
    )


class ToolPropertyValue(BaseModel):
    """
    A schema used to validate a single keyword argument for a tool.

    Attributes:
        type: The argument's type.

            The type is used to help the Agent generate valid arguments for the tool.

            For more information about types, see:
            https://json-schema.org/understanding-json-schema/reference/type.html#type-specific-keywords
        description: Description of what the argument is used for.

            This description is used to help the Agent generate sensible arguments for the tool.
            It is very important that this description is succinct, clear, and accurate.
        default: A default value for the argument if unspecified.
        examples: Example values demonstrating how the argument should look.

            This can be used to help the agent understand what a valid argument should look like.
    """
    type: Literal["string", "number", "integer", "boolean", "object", "array", "null",] = Field(
        ...,
    )
    description: str = Field(
        ...,
    )
    default: Optional[str] = Field(default=None)
    examples: Optional[List[str]] = Field(default=None)

    class Config(BaseModel.Config):
        title = "property"


class ToolArguments(BaseModel):
    """
    An object where each key is the name of a keyword argument and each value is a schema used to
    validate that property. Each schema must have a type and description, but can also have a
    default value and examples.

    For more information on how to define a valid property, visit
    https://json-schema.org/understanding-json-schema/reference/object.html

    Attributes:
        type: Type of argument. Currently only "object" is supported
        properties: An object where each key is the name of a keyword argument and each value is a
            schema used to validate that property. Each schema must have a type and description, but
            can also have a default value and examples.

            For more information on how to define a valid property, visit
            https://json-schema.org/understanding-json-schema/reference/object.html
    """
    type: Literal["object"] = Field(default="object")
    properties: Dict[str, ToolPropertyValue] = Field(default_factory=dict)


class Tool(BaseModel):
    """
    A tool is a function that the Agent has at its disposal. This schema is used to tell the Agent
    about the tools that are available to it.

    Attributes:
        name: Name of the tool.

            The tool name is the name the client wishes the Agent to use to refer to this
            function when it decides if it wants the user to use the tool or not. It must be
            unique amongst the set of tools provided in a single API call.
        description: Description of the tool.

            Because some queries are complex and may require multiple tools to complete, it is
            important to make these descriptions as informative as possible. If a tool is not
            being chosen when it should, it is common practice to tune the description of the
            tool to make it more apparent to the agent when the tool can be used effectively.
        arguments: An JSON Schema-compliant schema for the tool arguments. To describe a
            function that accepts no parameters, provide the value `{"type": "object",
            "properties": {}}`.

            For more information on how to define a valid JSON Schema object, visit
            https://json-schema.org/understanding-json-schema/reference/object.html
    """
    name: str = Field(..., pattern=r"^[a-zA-Z0-9_-]{1,64}$")
    description: str = Field(...)
    arguments: ToolArguments = Field(...)



from __future__ import annotations

from typing import Optional, Dict, Any, List

from pydantic import Field

from scale_egp.sdk.types.chunks import Chunk
from scale_egp.utils.model_utils import BaseModel


class KnowledgeBaseChunksRequest(BaseModel):
    chunk_id: Optional[str] = Field(None, description="Optional search by chunk_id")
    metadata_filters: Optional[Dict[str, Any]] = Field(
        None, description="Optional search by metadata fields"
    )


class KnowledgeBaseChunksResponse(BaseModel):
    chunks: List[Chunk] = Field(
        ..., description="List of chunks that match the chunk_id and metadata filters"
    )


class KnowledgeBaseQueryRequest(BaseModel):
    query: str = Field(
        description="The natural language query to be answered by referencing the data ingested "
                    "into the knowledge base"
    )
    top_k: int = Field(
        gt=0,
        description="Number of chunks to return. Must be greater than 0 if specified. If not "
                    "specified, all chunks will be returned.",
    )
    include_embeddings: bool = Field(
        default=True,
        description="Whether or not to include the embeddings for each chunk",
    )
    metadata_filters: Optional[Dict[str, Any]] = Field(
        None, description="Optional filter by metadata fields"
    )


class KnowledgeBaseQueryResponse(BaseModel):
    chunks: List[Chunk] = Field(
        description="An ordered list of the k most similar chunks and their similarity scores "
                    "from most to least similar"
    )



from __future__ import annotations

from typing import Dict, Optional, Union, List, Literal

from pydantic import Field

from scale_egp.utils.model_utils import Entity, BaseModel, RootEntity
from scale_egp.sdk.enums import QuestionType

class CategoricalChoice(BaseModel):
    """
    A choice for a categorical question.

    This is only used in `HUMAN` evaluation type to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        label: The text displayed to annotators for this choice.
        value: The value reported in the TestCaseResult for this question if this choice is
            selected.

            If users would like to track the improvement of a model over time, it is
            recommended to use a numeric value for this field.

            A string value can be used if no ordering is desired.
        audit_required: Whether selecting this choice will flag the test case result. Defaulted to false.
    """

    label: str
    value: Union[str, int, bool]
    audit_required: bool = Field(default=False)
    
    def to_dict(self):
        return {
            "label": self.label,
            "value": self.value,
            "audit_required": self.audit_required,
        }

# TODO <cathy-scale>: Turn Union[str, List[str]] into a CategoricalRule type when supporting more complex rules
CategoricalCondition = Dict[str, Union[str, List[str]]]
"""
A mapping from `question_id` to either the exact value that must be selected, or a list of
acceptable values. All key-value pairs in the mapping must be satisfied for the condition to be `True`.
For questions with `multi=True`, the selected values include at least one of the acceptable values.
"""

class BaseQuestion(Entity):
    """
    A base class for questions.

    Attributes:
        id: A unique id for this question. Each question_id will be represented as a
            key in the `result` field in the TestCaseResult after an annotation is completed.
            When examining how an application has performed on a specific question, users should
            look at the `result` field in the TestCaseResult for the key with the
            corresponding id.
        title: The text displayed to annotators for this question.
        prompt: The prompt displayed to annotators for this question.
        required: Whether this question is required. If
            `True`, the annotator must select at least one choice.
        conditions: Conditions that allow the question to be rendered.
            Example 1: `conditions=[{"accurate": "very", "complete": "yes"}]` means that the selected
            value for `accurate` must be `"very"` and the selected value for `complete` must be
            `"yes"`. Example 2: `conditions=[{"accurate": ["mostly", "very"]}, {"complete": "yes"}]`
            means that either the selected value for `accurate` must be `"mostly"` or `"very"`,
            or the selected value for `complete` must be `"yes"`.
        account_id: The account ID of the account that the question is associated with.
        created_at: The time the question was created.
        created_by_user_id: The user ID of the user who created the question.
    """
    id: str
    title: str
    prompt: str
    required: bool = Field(default=False)
    conditions: Optional[List[CategoricalCondition]]
    account_id: str
    created_at: str
    created_by_user_id: str


class CategoricalQuestion(BaseQuestion):
    """
    A categorical question.

    This is only used in `HUMAN` evaluation type to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        type: The type of the question.
        choices: The choices for the question.
        multi: Whether to allow multiple choices to be selected. If `True`, displays as a
            checkbox list. Otherwise, displays as a radio group.
        dropdown: Whether to display options as a dropdown list. If `True`, displays as a
            dropdown. Otherwise, displays as a radio group. Currently, dropdown cannot be true
            if multi is true.
    """

    type: Literal[QuestionType.CATEGORICAL] = QuestionType.CATEGORICAL.value
    choices: List[CategoricalChoice]
    multi: bool = Field(default=False)
    dropdown: Optional[bool] = Field(default=False)


class FreeTextQuestion(BaseQuestion):
    """
    A free text question.

    This is only used in `HUMAN` evaluation type to specify a choice for a  question
    that will be asked to users when they are evaluating generated outputs in the SGP annotation UI.

    Attributes:
        type: The type of the question.
    """

    type: Literal[QuestionType.FREE_TEXT] = QuestionType.FREE_TEXT.value


class Question(RootEntity):
    __root__: Union[CategoricalQuestion, FreeTextQuestion] = Field(
        ...,
        discriminator="type",
    )


class QuestionRequest(BaseModel):
    type: str
    title: str
    prompt: str
    account_id: str
    choices: Optional[List[dict]]
    multi: Optional[bool]
    dropdown: Optional[bool]
    required: Optional[bool]
    conditions: Optional[List[dict]]



# !!!!! DO NOT EVER CHANGE THIS FILE MANUALLY -- AUTOGENERATED by orm2pydantic.py !!!!!
# Generated from module egp_api_backend.server.internal.orm.orm_fine_tuning_jobs in the egp-api-backend package for use in the EGP python SDK
# To regenerate this file, run:
# scaleapi/packages/egp-api-backend/scripts/orm2pydantic.sh

from scale_egp.utils.model_utils import BaseModel, Entity
from pydantic import Field

from datetime import (datetime)
from scale_egp.sdk.models.fine_tuning_jobs import (FineTuningJobStatus, FineTuningJobVendorConfiguration, TrainingDatasetORMSchemaTypeEnum)
from scale_egp.utils.model_utils import (BaseModel)
from scale_egp.utils.strenum_compat import (StrEnum)
from typing import (Optional)
DataSource = StrEnum('DataSource', [('S3', 'S3'), ('SHARE_POINT', 'SharePoint'), ('LOCAL_FILE', 'LocalFile'), ('LOCAL_CHUNKS', 'LocalChunks'), ('GOOGLE_DRIVE', 'GoogleDrive'), ('AZURE_BLOB_STORAGE', 'AzureBlobStorage'), ('CONFLUENCE', 'Confluence')])
class TrainingDatasetRequest(BaseModel):
    name: str = Field(..., description="The name of the dataset")
    schema_type: TrainingDatasetORMSchemaTypeEnum = Field(..., description="The schema type of the dataset, currently only GENERATION is supported")
    data_source: DataSource = Field(..., description="The data source of the dataset, used to determine how to parse the location")
    account_id: str = Field(..., description="The ID of the account that owns the given entity.", can_patch=False)


class TrainingDataset(Entity):
    name: str = Field(..., description="The name of the dataset")
    schema_type: TrainingDatasetORMSchemaTypeEnum = Field(..., description="The schema type of the dataset, currently only GENERATION is supported")
    data_source: DataSource = Field(..., description="The data source of the dataset, used to determine how to parse the location")
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime = Field(..., description="The date and time when the entity was created in ISO format.")
    account_id: str = Field(..., description="The ID of the account that owns the given entity.", can_patch=False)
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")


class FineTuningJobRequest(BaseModel):
    base_model_id: Optional[str] = Field(None)
    vendor_configuration: Optional[FineTuningJobVendorConfiguration] = Field(None)
    fine_tuned_model_id: Optional[str] = Field(None)
    training_dataset_id: str
    validation_dataset_id: Optional[str] = Field(None)
    account_id: str = Field(..., description="The ID of the account that owns the given entity.", can_patch=False)


class FineTuningJob(Entity):
    base_model_id: Optional[str] = Field(None)
    vendor_configuration: Optional[FineTuningJobVendorConfiguration] = Field(None)
    fine_tuned_model_id: Optional[str] = Field(None)
    training_dataset_id: str
    validation_dataset_id: Optional[str] = Field(None)
    status: FineTuningJobStatus
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime = Field(..., description="The date and time when the entity was created in ISO format.")
    account_id: str = Field(..., description="The ID of the account that owns the given entity.", can_patch=False)
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")





from __future__ import annotations

from typing import List, Optional


from scale_egp.utils.model_utils import Entity, BaseModel
from scale_egp.sdk.types.questions import Question


class QuestionSet(Entity):
    """
    A data model representing a question set.

    Attributes:
        id: The ID of the question set
        name: The name of the question set
        created_at: The time the question set was created
        created_by_user_id: The user ID of the user who created the question set
        account_id: The account ID of the account that the question set is associated with
        questions: The questions in the question set
    """

    id: str
    name: str
    created_at: str
    created_by_user_id: str
    account_id: str
    questions: Optional[List[Question]]


class QuestionSetRequest(BaseModel):
    name: str
    account_id: str
    question_ids: List[str]




from datetime import datetime
from typing import Optional
from pydantic import Field
from scale_egp.utils.model_utils import BaseModel, Entity

class ModelGroup(Entity):
    """
    Entity for grouping models which are tied to the base model. E.g.: gpt-4 can be a group containing all gpt-4 fine-tuned models

    Attributes:
        id: The unique identifier of the entity.
        name: The name of the group
        description: Description of the group
    """

    name: str
    description: str
    id: str = Field(..., description="The unique identifier of the entity.")
    created_at: datetime = Field(
        ..., description="The date and time when the entity was created in ISO format."
    )
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")
    created_by_user_id: str = Field(..., description="The user who originally created the entity.")



class ModelGroupRequest(BaseModel):
    name: str
    description: Optional[str]
    account_id: str = Field(..., description="The ID of the account that owns the given entity.")



from __future__ import annotations

from datetime import datetime
from typing import Optional, List

from pydantic import Field

from scale_egp.sdk.enums import ArtifactSource
from scale_egp.sdk.types import Chunk
from scale_egp.utils.model_utils import BaseModel


class ChunksStatus(BaseModel):
    """
    A data model representing the status of the chunks in an artifact.

    Attributes:
        chunks_completed: Number of chunks uploaded successfully
        chunks_pending: Number of chunks awaiting upload
        chunks_failed: Number of chunks that failed upload
    """

    chunks_completed: int
    chunks_pending: int
    chunks_failed: int


class KnowledgeBaseArtifact(BaseModel):
    """
    A data model representing an artifact in a knowledge base.

    Attributes:
        artifact_id: Unique identifier for the artifact
        artifact_name: Friendly name for the artifact
        artifact_uri: Location (e.g. URI) of the artifact in the data source
        artifact_uri_public: Public Location (e.g. URI) of the artifact in the data source
        status: Status of the artifact
        status_reason: Reason for the artifact's status
        source: Data source of the artifact
        chunks_status: Number of chunks pending, completed, and failed
        updated_at: Timestamp at which the artifact was last updated
        chunks: List of chunks associated with the artifact
    """

    artifact_id: str
    artifact_name: str
    artifact_uri: str
    artifact_uri_public: Optional[str] = None
    status: str
    status_reason: Optional[str]
    source: ArtifactSource
    chunks_status: ChunksStatus
    updated_at: Optional[datetime] = None
    chunks: Optional[List[Chunk]] = None


class ListKnowledgeBaseArtifactsResponse(BaseModel):
    artifacts: List[KnowledgeBaseArtifact] = Field(..., description="List of artifacts.")



import hashlib
import logging
import re
import time
from typing import Optional, Dict, Any, List
from uuid import uuid4

import boto3
import os
import sys
import tarfile
import shutil

import botocore
import requests
import docker
import base64
import json

from urllib.parse import urlparse
from datetime import datetime
from argh import arg
from httpx import TimeoutException
from pydantic import BaseModel, Field

from scale_egp.cli.collections import GenericCommands
from scale_egp.exceptions import EGPException
from scale_egp.sdk.enums import EmbeddingModelName
from scale_egp.sdk.models.model_enums import ModelType
from scale_egp.sdk.types import ModelTemplateRequest, ModelTemplate, ModelInstanceRequest, ModelInstance, \
    ModelDeploymentRequest, ModelDeployment, ChunkToUpload, LocalChunksSourceConfig, EmbeddingRequest, \
    CompletionRequest, ParameterBindings

LOCAL_DATA_DIR = os.path.join(os.path.expanduser("~"), ".scale-egp", "aws-setup")
CACHE_DIR = os.path.join(LOCAL_DATA_DIR, "cache")
CACHE_UNTAR_DIR = os.path.join(CACHE_DIR, "untar")


# Set up logging to just print the string to stderr
formatter = logging.Formatter('%(message)s')
handler = logging.StreamHandler(sys.stderr)
handler.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(handler)

TEST_EMBEDDING_REQUEST = EmbeddingRequest(texts=["hello"])
TEST_COMPLETION_REQUEST = {"prompts": ["hello"]}


def debug_print(*args, **kwargs):
    logger.info(*args, **kwargs)


def error_print(*args, **kwargs):
    logger.error(*args, **kwargs)


EMBEDDING_MODEL_FINE_TUNED = "finetuned"
EMBEDDING_MODEL_OPEN_SOURCE = "opensource"


class EmbeddingConfig(BaseModel):
    embedding_model: str


class KnowledgeBaseData(BaseModel):
    embedding_config: EmbeddingConfig
    chunks: List[ChunkToUpload]


class AppConfigInput(BaseModel):
    knowledge_bases: Dict[str, KnowledgeBaseData]


class AwsVpcSetupCommands(GenericCommands):
    command_group_name = "aws-setup"
    command_group_title = "AWS VPC Setup Commands"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._egp_client_val = None
        self._s3_util_client_val = None
        self._ecr_util_client_val = None

    @property
    def _egp_client(self):
        if self._egp_client_val is None:
            self._egp_client_val = self._client_factory.get_client()
        return self._egp_client_val

    @property
    def _s3_util_client(self):
        if self._s3_util_client_val is None:
            try:
                self._s3_util_client_val = S3UtilClient(boto3.client("s3"))
            except Exception as e:
                debug_print(f"Error creating S3 client: {e}")
                debug_print("Check your AWS credentials and try again.")
                sys.exit(1)
        return self._s3_util_client_val

    @property
    def _ecr_util_client(self):
        if self._ecr_util_client_val is None:
            try:
                ecr_client = boto3.client("ecr")
            except Exception as e:
                debug_print(f"Error creating ECR client: {e}")
                debug_print("Check your AWS credentials and try again.")
                sys.exit(1)

            try:
                docker_client = docker.from_env()
            except Exception as e:
                debug_print(f"Error creating Docker client: {e}")
                debug_print("Check your Docker installation / setup (like environment variables) and try again.")
                sys.exit(1)

            ecr_url = None
            try:
                ecr_credentials = ecr_client.get_authorization_token()["authorizationData"][0]
                ecr_username = "AWS"
                ecr_password = base64.b64decode(ecr_credentials["authorizationToken"]).decode().split(":")[1]
                ecr_url = ecr_credentials["proxyEndpoint"].replace("https://", "")

                docker_client.login(username=ecr_username, password=ecr_password, registry=ecr_url, reauth=True)
            except Exception as e:
                debug_print(f"Error logging into ECR: {e}")
                debug_print("Check your AWS credentials and try again.")
                if ecr_url:
                    debug_print(f"If your credentials are expired, try running `docker logout {ecr_url}` and try again")
                sys.exit(1)

            self._ecr_util_client_val = EcrUtilClient(ecr_client, docker_client)
        return self._ecr_util_client_val

    @arg("--model-config-src-uri", required=False, help="URI pointing to the model config file (currently only supports http(s) URIs). One of src-uri or src-file must be provided")
    @arg("--model-config-src-file", required=False, help="File path for a config file. One of src-uri or src-file must be provided")
    @arg("--model-weights-src-uri", required=False, help="URI pointing to the model weights file (currently only supports http(s) URIs). Required if model weights do not exist in S3")
    @arg("--model-weights-dest-uri", help="The model weight destination S3 URI (bucket and key) (key prefix if unpacking)")
    @arg("--model-image-src-cr-uri", required=False, help="Registry / repository URI pointing to the model image. If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--model-image-src-tar-uri", required=False, help="URI pointing to the model image tar file (currently only supports http(s) URIs). If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--model-image-dest-uri", required=True, help="The model image destination ECR URI")
    @arg("--debug", required=False, help="Enable debug logging", default=False)
    def create_model_deployment(self, *, model_config_src_uri, model_config_src_file, model_weights_src_uri, model_weights_dest_uri, model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri, debug):
        """
        Requirements: AWS CLI, Docker
        Must have AWS credentials configured with permissions to S3 and ECR.

        This command creates a model deployment in the Scale EGP given a model
        configuration, model weights, and model image. Scale will provide you
        with these in the form of https URLs for the model weights and image,
        and either a JSON file or a URL for the model configuration.

        You will need to provide the installation locations of the model weights
        and image. In the default Scale EGP configuration, the model weights
        should be stored in the bucket scale-egp-<workspace>-models and the image
        should be stored in the ECR repository scale-egp-<workspace-id>-launch/inference
        or vllm based on the model type.

        Here is an example of how to use this command:

        export EGP_ENDPOINT_URL=https://api.<workspace-id>.workspace.egp.scale.com/
        export EGP_ACCOUNT_ID=<account-id>
        export EGP_API_KEY=<api-key>
        scale-egp aws-setup create-model-deployment \\
            --model-weights-src-uri "<scale-provided-model-weights-url>" \\
            --model-weights-dest-uri "s3://scale-egp-<workspace-id>-models/my-model-weights" \\
            --model-image-src-tar-uri "scale-provided-model-image-url>" \\
            --model-image-dest-uri "<aws-account-id>.dkr.ecr.us-west-2.amazonaws.com/vllm:1.0.0" \\
            --model-config-src-uri "<scale-provided-model-config-url>"

        If you already have the model weights and image in S3 and ECR, you can skip
        providing the sources. The command would then look like this:

        scale-egp aws-setup create-model-deployment \\
            --model-weights-dest-uri "s3://scale-egp-<workspace-id>-models/my-model-weights" \\
            --model-image-dest-uri "<aws-account-id>.dkr.ecr.us-west-2.amazonaws.com/vllm:1.0.0" \\
            --model-config-src-uri "<scale-provided-model-config-url>"

        The model weights and image files are cached locally to avoid repeated downloads.
        The cache is located at ~/.scale-egp/aws-setup/ and can be cleared with the command:

        scale-egp aws-setup clear-cache
        """
        if debug:
            logger.setLevel(logging.INFO)

        self._create_model_deployment_logic(model_config_src_uri, model_config_src_file, model_weights_src_uri, model_weights_dest_uri, model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri)

    def _create_model_deployment_logic(self, model_config_src_uri, model_config_src_file, model_weights_src_uri, model_weights_dest_uri, model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri):
        model_template_request, model_instance_request, model_deployment_request = self._load_model_config(model_config_src_uri, model_config_src_file)
        should_unpack_weights = model_template_request.model_type in [ModelType.COMPLETION]
        model_weights_bucket, model_weights_key = split_s3_uri(model_weights_dest_uri)

        should_install_model_weights = self._check_model_weights_options(model_weights_src_uri, model_weights_bucket, model_weights_key, should_unpack_weights)
        should_install_model_image = self._check_model_image_options(model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri)

        if should_install_model_weights:
            self._install_model_weights(model_weights_src_uri, model_weights_bucket, model_weights_key, should_unpack_weights)
        else:
            debug_print(f"Model weights already exist in s3://{model_weights_bucket}/{model_weights_key}")
            debug_print(f"Skipping installation of model weights.")

        if should_install_model_image:
            self._install_model_image(model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri)
        else:
            debug_print(f"Model image already exists in {model_image_dest_uri}")
            debug_print(f"Skipping installation of model image.")

        model_instance, model_deployment = self._deploy_model(model_weights_bucket, model_weights_key, model_image_dest_uri, model_template_request, model_instance_request, model_deployment_request)

        self._check_model_deployed(model_instance, model_deployment, model_template_request.model_type)

        debug_print("\n\n\n")
        print("Model instance ID:", model_instance.id)
        print("Model deployment ID:", model_deployment.id)

        return model_instance, model_deployment

    @arg("--os-embedding-instance-id", required=True, help="The model instance ID of the OPEN SOURCE embedding model to use (NOTE: this is the model __instance__)")
    @arg("--os-embedding-deployment-id", required=True, help="The deployment ID of the OPEN SOURCE embedding model to use (NOTE: this is the model __deployment__)")
    @arg("--ft-embedding-instance-id", required=True, help="The model instance ID of the FINE TUNED embedding model to use (NOTE: this is the model __instance__)")
    @arg("--ft-embedding-deployment-id", required=True, help="The deployment ID of the FINE TUNED embedding model to use (NOTE: this is the model __deployment__)")
    @arg("--llm-instance-id", required=True, help="The model instance ID of the LLM model to use (NOTE: this is the model __instance__)")
    @arg("--llm-deployment-id", required=True, help="The deployment ID of the LLM model to use (NOTE: this is the model __deployment__)")
    @arg("--app-data-uri", required=True, help="URI pointing to a file containing applications knowledge base data (currently only supports http(s) URIs)")
    @arg("--debug", required=False, help="Enable debug logging", default=False)
    def create_app_configs(self, *, os_embedding_instance_id, os_embedding_deployment_id, ft_embedding_instance_id, ft_embedding_deployment_id, llm_instance_id, llm_deployment_id, app_data_uri, debug):
        """
        Creates application configurations for Scale MLE developed apps,
        by creating knowledge bases and prompt generators for each app.
        Currently supports the creation of Text2SQL applications.

        The model instance and deployment IDs can be obtained by running
        the create-model-deployment subcommand, given the Scale provided
        configuration. The app data URI should point to a JSON file provided
        by Scale.
        """
        if debug:
            logger.setLevel(logging.INFO)

        self._create_app_configs_logic(
            os_embedding_instance_id,
            os_embedding_deployment_id,
            ft_embedding_instance_id,
            ft_embedding_deployment_id,
            llm_instance_id,
            llm_deployment_id,
            app_data_uri
        )

    def _create_app_configs_logic(self, os_embedding_instance_id, os_embedding_deployment_id, ft_embedding_instance_id, ft_embedding_deployment_id, llm_instance_id, llm_deployment_id, app_data_uri):

        applications = self._load_app_data_config(app_data_uri)

        env_vars = []
        for name, app_config_input in applications.items():
            ret_env_vars = self._load_knowledge_bases_and_create_app(name, app_config_input, os_embedding_deployment_id, ft_embedding_deployment_id)
            env_vars.extend(ret_env_vars)

        debug_print("\n\n\n")

        print(f"MODEL_ID={llm_instance_id}")
        print(f"MODEL_DEPLOYMENT_ID={llm_deployment_id}")
        for env_var in env_vars:
            print(env_var)

    @arg("--os-emb-model-config-src-uri", required=False,
         help="OPEN SOURCE EMBEDDING: URI pointing to the model config file (currently only supports http(s) URIs). One of src-uri or src-file must be provided")
    @arg("--os-emb-model-config-src-file", required=False,
         help="OPEN SOURCE EMBEDDING: File path for a config file. One of src-uri or src-file must be provided")
    @arg("--os-emb-model-weights-src-uri", required=False,
         help="OPEN SOURCE EMBEDDING: URI pointing to the model weights file (currently only supports http(s) URIs). Required if model weights do not exist in S3")
    @arg("--os-emb-model-weights-dest-uri",
         help="OPEN SOURCE EMBEDDING: The model weight destination S3 URI (bucket and key) (key prefix if unpacking)")
    @arg("--os-emb-model-image-src-cr-uri", required=False,
         help="OPEN SOURCE EMBEDDING: Registry / repository URI pointing to the model image. If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--os-emb-model-image-src-tar-uri", required=False,
         help="OPEN SOURCE EMBEDDING: URI pointing to the model image tar file (currently only supports http(s) URIs). If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--os-emb-model-image-dest-uri", required=True, help="OPEN SOURCE EMBEDDING: The model image destination ECR URI")
    @arg("--ft-emb-model-config-src-uri", required=False,
         help="FINE TUNED EMBEDDING: URI pointing to the model config file (currently only supports http(s) URIs). One of src-uri or src-file must be provided")
    @arg("--ft-emb-model-config-src-file", required=False,
         help="FINE TUNED EMBEDDING: File path for a config file. One of src-uri or src-file must be provided")
    @arg("--ft-emb-model-weights-src-uri", required=False,
         help="FINE TUNED EMBEDDING: URI pointing to the model weights file (currently only supports http(s) URIs). Required if model weights do not exist in S3")
    @arg("--ft-emb-model-weights-dest-uri",
         help="FINE TUNED EMBEDDING: The model weight destination S3 URI (bucket and key) (key prefix if unpacking)")
    @arg("--ft-emb-model-image-src-cr-uri", required=False,
         help="FINE TUNED EMBEDDING: Registry / repository URI pointing to the model image. If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--ft-emb-model-image-src-tar-uri", required=False,
         help="FINE TUNED EMBEDDING: URI pointing to the model image tar file (currently only supports http(s) URIs). If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--ft-emb-model-image-dest-uri", required=True,
         help="FINE TUNED EMBEDDING: The model image destination ECR URI")
    @arg("--llm-model-config-src-uri", required=False,
         help="LLM: URI pointing to the model config file (currently only supports http(s) URIs). One of src-uri or src-file must be provided")
    @arg("--llm-model-config-src-file", required=False,
         help="LLM: File path for a config file. One of src-uri or src-file must be provided")
    @arg("--llm-model-weights-src-uri", required=False,
         help="LLM: URI pointing to the model weights file (currently only supports http(s) URIs). Required if model weights do not exist in S3")
    @arg("--llm-model-weights-dest-uri",
         help="LLM: The model weight destination S3 URI (bucket and key) (key prefix if unpacking)")
    @arg("--llm-model-image-src-cr-uri", required=False,
         help="LLM: Registry / repository URI pointing to the model image. If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--llm-model-image-src-tar-uri", required=False,
         help="LLM: URI pointing to the model image tar file (currently only supports http(s) URIs). If image does not exist in ECR, one of cr-uri or tar-uri must be provided")
    @arg("--llm-model-image-dest-uri", required=True, help="LLM: The model image destination ECR URI")
    @arg("--app-data-uri", required=True,
         help="URI pointing to a file containing applications knowledge base data (currently only supports http(s) URIs)")
    @arg("--debug", required=False, help="Enable debug logging", default=False)
    def create_models_and_app_configs(self, *,
        os_emb_model_config_src_uri,
        os_emb_model_config_src_file,
        os_emb_model_weights_src_uri,
        os_emb_model_weights_dest_uri,
        os_emb_model_image_src_cr_uri,
        os_emb_model_image_src_tar_uri,
        os_emb_model_image_dest_uri,
        ft_emb_model_config_src_uri,
        ft_emb_model_config_src_file,
        ft_emb_model_weights_src_uri,
        ft_emb_model_weights_dest_uri,
        ft_emb_model_image_src_cr_uri,
        ft_emb_model_image_src_tar_uri,
        ft_emb_model_image_dest_uri,
        llm_model_config_src_uri,
        llm_model_config_src_file,
        llm_model_weights_src_uri,
        llm_model_weights_dest_uri,
        llm_model_image_src_cr_uri,
        llm_model_image_src_tar_uri,
        llm_model_image_dest_uri,
        app_data_uri,
        debug
    ):
        if debug:
            logger.setLevel(logging.INFO)

        os_embedding_instance, os_embedding_deployment = self._create_model_deployment_logic(
            os_emb_model_config_src_uri,
            os_emb_model_config_src_file,
            os_emb_model_weights_src_uri,
            os_emb_model_weights_dest_uri,
            os_emb_model_image_src_cr_uri,
            os_emb_model_image_src_tar_uri,
            os_emb_model_image_dest_uri
        )

        ft_embedding_instance, ft_embedding_deployment = self._create_model_deployment_logic(
            ft_emb_model_config_src_uri,
            ft_emb_model_config_src_file,
            ft_emb_model_weights_src_uri,
            ft_emb_model_weights_dest_uri,
            ft_emb_model_image_src_cr_uri,
            ft_emb_model_image_src_tar_uri,
            ft_emb_model_image_dest_uri
        )

        llm_instance, llm_deployment = self._create_model_deployment_logic(
            llm_model_config_src_uri,
            llm_model_config_src_file,
            llm_model_weights_src_uri,
            llm_model_weights_dest_uri,
            llm_model_image_src_cr_uri,
            llm_model_image_src_tar_uri,
            llm_model_image_dest_uri
        )

        self._create_app_configs_logic(
            os_embedding_instance.id,
            os_embedding_deployment.id,
            ft_embedding_instance.id,
            ft_embedding_deployment.id,
            llm_instance.id,
            llm_deployment.id,
            app_data_uri
        )

    @arg("--debug", required=False, help="Enable debug logging", default=False)
    def clear_cache(self, *, debug=False):
        """
        This command clears the local cache of downloaded files.
        """
        if debug:
            logger.setLevel(logging.INFO)

        try:
            shutil.rmtree(LOCAL_DATA_DIR)
            print(f"Cleared local cache at {LOCAL_DATA_DIR}.")
        except Exception as e:
            error_print(f"Error clearing local cache: {e}")
            error_print("Check the directory permissions and try again.")
            sys.exit(1)


    def _parse_model_configs(self, model_config):
        model_template_config = model_config["model_template"]
        model_template_config["account_id"] = self._egp_client.account_id
        model_template_config["name"] = ""
        model_template_config["vendor_configuration"]["bundle_config"]["registry"] = ""
        model_template_config["vendor_configuration"]["bundle_config"]["image"] = ""
        model_template_config["vendor_configuration"]["bundle_config"]["tag"] = ""

        model_instance_config = model_config["model_instance"]
        model_instance_config["account_id"] = self._egp_client.account_id
        model_instance_config["name"] = ""
        model_instance_config["model_template_id"] = ""

        model_deploy_config = model_config["model_deployment"]
        model_deploy_config["account_id"] = self._egp_client.account_id
        model_deploy_config["name"] = ""

        return ModelTemplateRequest(**model_template_config), ModelInstanceRequest(**model_instance_config), ModelDeploymentRequest(**model_deploy_config)

    def _load_model_config(self, model_config_src_uri, model_config_src_file):
        if model_config_src_uri is None and model_config_src_file is None:
            error_print("Either model-config-src-uri or model-config-src-file must be provided.")
            sys.exit(1)

        if model_config_src_file is None:
            model_config_src_file = download_uri(model_config_src_uri)

        try:
            with open(model_config_src_file, "r") as f:
                model_config_contents = f.read()
            model_config = json.loads(model_config_contents)
        except Exception as e:
            error_print(f"Error reading model config from {model_config_src_file}: {e}")
            error_print("Check the model config source and try again.")
            sys.exit(1)

        try:
            return self._parse_model_configs(model_config)
        except KeyError as e:
            error_print(f"Error parsing model request from model config: missing or erroneous key {e}")
            error_print("Check the model config source is downloading correctly.")
            error_print("Reach out to Scale for further help.")
            sys.exit(1)
        except Exception as e:
            error_print(f"Error parsing model request from model config: {e}")
            error_print("Check the model config source and try again.")
            error_print("Reach out to Scale for further help.")
            sys.exit(1)

    def _check_model_weights_options(self, model_weights_src_uri, model_weights_dest_bucket, model_weights_dest_key, should_unpack):
        should_install_model_weights = True

        if should_unpack:
            if '.' in model_weights_dest_key:
                error_print("For the configured model type, the model weights should be stored in an S3 directory. Fix the destination key (remove file extension) and try again.")
                sys.exit(1)
            if self._s3_util_client.does_directory_exists(model_weights_dest_bucket, model_weights_dest_key):
                should_install_model_weights = False
        elif not should_unpack:
            if not model_weights_dest_key.endswith('.tar'):
                error_print("For the configured model type, the model weights should be stored in a tar file. Fix the destination key (add .tar extension) and try again.")
                sys.exit(1)
            if self._s3_util_client.does_file_exist(model_weights_dest_bucket, model_weights_dest_key):
                should_install_model_weights = False

        if should_install_model_weights and model_weights_src_uri is None:
            error_print("Parameter --model-weights-src-uri is required if model weights do not exist in S3.")
            sys.exit(1)

        return should_install_model_weights

    def _install_model_weights(self, model_weights_src_uri, destination_bucket, destination_key, should_unpack=False):
        debug_print("")
        debug_print(f"####################################################################")
        debug_print(f"####################  Installing model weights  ####################")
        debug_print(f"####################################################################")

        if should_unpack:
            model_weight_unpacked_dir = download_and_untar_uri(model_weights_src_uri)
            self._s3_util_client.copy_directory_to_s3(model_weight_unpacked_dir, destination_bucket, destination_key)
        else:
            model_weight_tar_file_path = download_uri(model_weights_src_uri)
            self._s3_util_client.copy_file_to_s3(model_weight_tar_file_path, destination_bucket, destination_key)

        debug_print(f"Model weights mirrored to s3://{destination_bucket}/{destination_key} successfully.")

    def _check_model_image_options(self, model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri):
        should_install_model_image = True
        if self._ecr_util_client.image_exists(model_image_dest_uri):
            should_install_model_image = False

        if should_install_model_image and model_image_src_cr_uri is None and model_image_src_tar_uri is None:
            error_print("Parameter --model-image-src-cr-uri or --model-image-src-tar-uri is required if model image does not exist in ECR.")
            sys.exit(1)

        return should_install_model_image

    def _install_model_image(self, model_image_src_cr_uri, model_image_src_tar_uri, model_image_dest_uri):
        debug_print("")
        debug_print(f"####################################################################")
        debug_print(f"####################  Installing model image  ######################")
        debug_print(f"####################################################################")

        if model_image_src_cr_uri is None and model_image_src_tar_uri is None:
            error_print("Either model-image-src-cr-uri or model-image-src-tar-uri must be provided.")
            sys.exit(1)

        if model_image_src_cr_uri is not None:
            self._ecr_util_client.mirror_image(model_image_src_cr_uri, model_image_dest_uri)
        else:
            image_tar_file_path = download_uri(model_image_src_tar_uri)
            self._ecr_util_client.upload_image_from_tar(image_tar_file_path, model_image_dest_uri)

        debug_print(f"Model image mirrored to {model_image_dest_uri} successfully.")

    def _deploy_model(self, model_weights_bucket, model_weights_key, model_image_uri, model_template_request, model_instance_request, model_deployment_request):
        debug_print("")
        debug_print(f"####################################################################")
        debug_print(f"#######################  Deploying model  ##########################")
        debug_print(f"####################################################################")

        registry, repository, tag = get_registry_repository_tag(model_image_uri)
        date_suffix = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        model_name = f"{repository}_{tag}__{model_weights_key}__{date_suffix}"

        model_template_request.name = model_name
        model_template_request.vendor_configuration.bundle_config.registry = registry
        model_template_request.vendor_configuration.bundle_config.image = repository
        model_template_request.vendor_configuration.bundle_config.tag = tag

        debug_print(f"Creating model template {model_name}...")
        try:
            model_template_collection = self._egp_client.model_templates()
            model_template_response = model_template_collection._post(model_template_collection._sub_path, model_template_request)
            if model_template_response.status_code != 200:
                raise Exception(f"Request error: {model_template_response.json()}")
            model_template_response_dict = model_template_response.json()
            if not isinstance(model_template_response_dict, dict):
                raise Exception(f"Unexpected response: {model_template_response_dict}")
            model_template = ModelTemplate(**model_template_response_dict)
        except Exception as e:
            error_print(f"Error creating model template: {e}")
            error_print("Check your API key and other API configuration.")
            error_print("Reach out to Scale for further help.")
            sys.exit(1)
        debug_print(f"Model template {model_name} created successfully with ID {model_template.id}.")

        model_instance_name = f"{model_name}___instance"

        model_instance_request.name = model_instance_name
        model_instance_request.model_template_id = model_template.id

        debug_print(f"Creating model instance {model_instance_name}...")
        try:
            model_instance_collection = self._egp_client.models()
            model_instance_response = model_instance_collection._post(model_instance_collection._sub_path, model_instance_request)
            if model_instance_response.status_code != 200:
                raise Exception(f"Request error: {model_instance_response.json()}")
            model_instance_response_dict = model_instance_response.json()
            if not isinstance(model_instance_response_dict, dict):
                raise Exception(f"Unexpected response: {model_instance_response_dict}")
            model_instance = ModelInstance(**model_instance_response_dict)
        except Exception as e:
            error_print(f"Error creating model instance: {e}")
            error_print("Check your API key and other API configuration.")
            error_print("Reach out to Scale for further help.")
            sys.exit(1)
        debug_print(f"Model instance {model_instance_name} created successfully with ID {model_instance.id}.")

        model_deployment_name = f"{model_name}___deployment"

        model_deployment_request.name = model_deployment_name
        model_deployment_request.model_creation_parameters = {
            "MODEL_BUCKET": model_weights_bucket,
            "MODEL_PATH": model_weights_key
        }

        debug_print(f"Deploying model {model_deployment_name}...")
        model_deployment_collection = self._egp_client.models().deployments()
        path = model_deployment_collection._sub_path.format(model_id=model_instance.id)

        deployment_try_count = 0
        while True:
            try:
                model_deployment_response = model_deployment_collection._post(path, model_deployment_request)
                if model_deployment_response.status_code != 200:
                    raise Exception(f"Request error: {model_deployment_response.json()}")
                model_deployment_response_dict = model_deployment_response.json()
                if not isinstance(model_deployment_response_dict, dict):
                    raise Exception(f"Unexpected response: {model_deployment_response_dict}")
                model_deployment = ModelDeployment(**model_deployment_response_dict)
                break
            except Exception as e:
                deployment_try_count += 1
                if deployment_try_count > 3:
                    error_print(f"Error deploying model: {e}")
                    error_print("Check your API key and other API configuration.")
                    error_print("Reach out to Scale for further help.")
                    sys.exit(1)
                debug_print(f"Error deploying model: {e}")
                debug_print(f"Retrying deployment {deployment_try_count}...")
                time.sleep(10)
        debug_print(model_deployment)
        debug_print(f"Model {model_deployment_name} deployed successfully with ID {model_deployment.id}.")

        poll_count = 0
        internal_service_error_count = 0
        debug_print(f"Deployment Status: {model_deployment.status}")
        while model_deployment.status != "READY":
            poll_count += 1
            time.sleep(10)

            try:
                model_deployment = self._egp_client.models().deployments().get(id=model_deployment.id, model=model_instance)
            except EGPException as e:
                if 500 <= e.code < 600:
                    internal_service_error_count += 1
                    debug_print(f"Internal service error: {e}")
                    debug_print(f"Internal service error count: {internal_service_error_count}")
                    if internal_service_error_count > 3:
                        debug_print(f"Error testing model deployment {model_deployment.id}, under instance {model_instance.id}: {e}")
                        debug_print("Check the model deployment and instance configurations.")
                        debug_print("Reach out to Scale for further help.")
                        sys.exit(1)
                    time.sleep(10)
                    continue
                else:
                    error_print(f"Error testing model deployment {model_deployment.id}, under instance {model_instance.id}: {e}")
                    error_print("Check the model deployment and instance configurations.")
                    error_print("Reach out to Scale for further help.")
                    sys.exit(1)
            except Exception as e:
                debug_print(e.__class__)
                debug_print(e.__class__.__name__)

            debug_print(f"Waiting on Model Deployment, Status: {model_deployment.status}")
            debug_print(f"Poll count: {poll_count}")

            if poll_count > 12:
                raise TimeoutError("Model deployment took too long to complete. Exiting")

        debug_print(f"Model deployment {model_deployment_name} is ready.")

        return model_instance, model_deployment

    def _load_app_data_config(self, app_data_uri):
        try:
            app_data_file_path = download_uri(app_data_uri)
        except Exception as e:
            error_print(f"Error downloading knowledge base data from {app_data_uri}: {e}")
            error_print("Check the knowledge base data source and try again.")
            sys.exit(1)

        try:
            with open(app_data_file_path, "r") as f:
                app_data = json.load(f)

            if not isinstance(app_data, dict):
                raise ValueError("App data must be a dictionary.")

            parsed_app_data = {key: AppConfigInput(**value) for key, value in app_data.items()}
            return parsed_app_data
        except Exception as e:
            error_print(f"Error parsing app data: {e}")
            error_print("Check the knowledge base data source and try again.")
            sys.exit(1)

    def _load_knowledge_bases_and_create_app(self, name: str, app_kb_input: AppConfigInput, os_embedding_deployment_id, ft_embedding_deployment_id):
        debug_print("")
        debug_print(f"####################################################################")
        debug_print(f"######################  Creating Application  ######################")
        debug_print(f"####################################################################")

        debug_print(f"Creating app {name}...")

        env_settings = []
        time_str = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        for kb_key, knowledge_base_data in app_kb_input.knowledge_bases.items():
            embedding_deployment_id = os_embedding_deployment_id
            if knowledge_base_data.embedding_config.embedding_model == EMBEDDING_MODEL_FINE_TUNED:
                embedding_deployment_id = ft_embedding_deployment_id

            kb_id = self._load_knowledge_base(
                f"{name}-{kb_key}-{embedding_deployment_id}--{time_str}",
                embedding_deployment_id,
                knowledge_base_data
            )
            env_key = f"{name.upper()}_{kb_key.upper()}_KNOWLEDGE_BASE_ID"
            env_setting = f"{env_key}={kb_id}"
            env_settings.append(env_setting)

        return env_settings

    def _load_knowledge_base(self, name, embedding_deployment_id, knowledge_base_data: KnowledgeBaseData):
        debug_print(f"Creating knowledge base {name}...")
        try:
            knowledge_base = self._egp_client.knowledge_bases().create(
                name=name,
                model_deployment_id=embedding_deployment_id
            )
        except Exception as e:
            error_print(f"Error creating knowledge base {name}: {e}")
            error_print("Check your API key and other API configuration.")
            error_print("Reach out to Scale for further help.")
            sys.exit(1)
        debug_print(f"Knowledge base {name} created successfully with ID {knowledge_base.knowledge_base_id}.")

        debug_print(f"Creating upload for knowledge base {name}...")
        try:
            upload = self._egp_client.knowledge_bases().uploads().create_local_upload(
                knowledge_base=knowledge_base,
                chunks=knowledge_base_data.chunks,
                data_source_config=LocalChunksSourceConfig(
                    artifact_name=f"{name}_upload",
                    artifact_uri=f"{knowledge_base.knowledge_base_name}_{uuid4()}",
                    deduplication_strategy="Overwrite"
                )
            )
        except Exception as e:
            error_print(f"Error uploading knowledge base {name}: {e}")
            error_print("Check your API key and other API configuration.")
            error_print("Reach out to Scale for further help.")
            sys.exit(1)
        debug_print(f"Knowledge base {name} upload started with ID {upload.upload_id}.")

        poll_count = 0
        while upload.status != "Completed":
            time.sleep(10)
            poll_count += 1
            try:
                upload = self._egp_client.knowledge_bases().uploads().get(
                    id=upload.upload_id,
                    knowledge_base=knowledge_base
                )
            except Exception as e:
                error_print(f"Error polling knowledge base {name} upload: {e}")
                error_print("Check your API key and other API configuration.")
                error_print("Reach out to Scale for further help.")
                sys.exit(1)
            debug_print(f"Waiting on Knowledge Base Upload, Status: {upload.status}")
            debug_print(f"Poll count: {poll_count}")
            debug_print(f"Status reason: {upload.status_reason}")
            debug_print(f"Artifact statuses: {upload.artifacts_status}")

        debug_print(f"Knowledge base {name} upload completed successfully.")

        return knowledge_base.knowledge_base_id

    def _check_model_deployed(self, model_instance, model_deployment, model_type):
        request = TEST_EMBEDDING_REQUEST if model_type == ModelType.EMBEDDING else TEST_COMPLETION_REQUEST
        poll_count = 0
        timeout = 30
        internal_service_error_count = 0
        debug_print(f"Testing model deployment {model_deployment.id}, under instance {model_instance.id}...")
        while True:
            debug_print(f"Waited {poll_count * timeout} seconds for model to be ready")
            poll_count += 1
            try:
                result = self._egp_client.models().deployments().execute(
                    model_deployment.id,
                    model_instance,
                    request,
                    timeout=timeout
                )
                break
            except TimeoutException as e:
                continue
            except EGPException as e:
                if 500 <= e.code < 600:
                    internal_service_error_count += 1
                    debug_print(f"Internal service error: {e}")
                    debug_print(f"Internal service error count: {internal_service_error_count}")
                    if internal_service_error_count > 3:
                        debug_print(f"Error testing model deployment {model_deployment.id}, under instance {model_instance.id}: {e}")
                        debug_print("Check the model deployment and instance configurations.")
                        debug_print("Reach out to Scale for further help.")
                        sys.exit(1)
                    time.sleep(10)
                    continue
                else:
                    error_print(f"Error testing model deployment {model_deployment.id}, under instance {model_instance.id}: {e}")
                    error_print("Check the model deployment and instance configurations.")
                    error_print("Reach out to Scale for further help.")
                    sys.exit(1)
            except Exception as e:
                error_print(f"Error testing model deployment {model_deployment.id}, under instance {model_instance.id}: {e}")
                error_print("Reach out to Scale for further help.")
                sys.exit(1)
        print("Model deployment is ready.")


class S3UtilClient:
    def __init__(self, s3_client: boto3.session.Session.client):
        self.s3_client = s3_client

    def _list_objects_v2(self, Bucket=None, Prefix=None):
        try:
            return self.s3_client.list_objects_v2(Bucket=Bucket, Prefix=Prefix)
        except Exception as e:
            error_print(f"Error listing objects in s3://{Bucket}/{Prefix}: {e}")
            error_print("Check your AWS credentials or role access and try again.")
            sys.exit(1)

    def _upload_file(self, file_path, bucket, key):
        try:
            self.s3_client.upload_file(file_path, bucket, key)
        except Exception as e:
            error_print(f"Error uploading file {file_path} to s3://{bucket}/{key}: {e}")
            error_print("Check your AWS credentials or role access and try again.")
            sys.exit(1)

    def does_file_exist(self, bucket, file_key):
        try:
            try:
                self.s3_client.head_object(Bucket=bucket, Key=file_key)
                return True
            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == '404':
                    return False
                raise e
        except Exception as e:
            error_print(f"Error checking if object s3://{bucket}/{file_key} exists: {e}")
            error_print("Check your AWS credentials or role access and try again.")
            sys.exit(1)

    def does_directory_exists(self, bucket, key):
        if not key.endswith('/'):
            key = f"{key}/"

        response = self._list_objects_v2(Bucket=bucket, Prefix=key)
        return 'Contents' in response

    def copy_directory_to_s3(self, dir_path, destination_bucket, destination_key):
        debug_print(f"Uploading contents of {dir_path} to s3://{destination_bucket}/{destination_key} ...")
        for root, dirs, files in os.walk(dir_path):
            for file in files:
                # Skip hidden files (created by tar.extractall)
                if file.startswith('.'):
                    continue

                local_file_path = os.path.join(root, file)
                relative_path = os.path.relpath(local_file_path, dir_path)
                s3_key = os.path.join(destination_key, relative_path)
                self._upload_file(local_file_path, destination_bucket, s3_key)
        debug_print(f"Finished uploading to s3.")

    def copy_file_to_s3(self, file_path, destination_bucket, destination_key):
        debug_print(f"Uploading {file_path} to s3://{destination_bucket}/{destination_key} ...")
        self._upload_file(file_path, destination_bucket, destination_key)
        debug_print(f"Finished uploading to s3")


class EcrUtilClient:

    def __init__(self, ecr_client, docker_client):
        self.ecr_client = ecr_client
        self.docker_client = docker_client

    def image_exists(self, image_uri):
        _, repository, tag = get_registry_repository_tag(image_uri)
        try:
            response = self.ecr_client.describe_images(
                repositoryName=repository,
                imageIds=[
                    {
                        'imageTag': tag
                    }
                ]
            )
            return len(response['imageDetails']) > 0
        except self.ecr_client.exceptions.ImageNotFoundException:
            return False
        except self.ecr_client.exceptions.RepositoryNotFoundException:
            error_print(f"Repository {repository} not found.")
            error_print("Check the repository name or create it in the ECR and try again.")
            sys.exit(1)
        except Exception as e:
            error_print(f"Error checking if image {image_uri} exists: {e}")
            error_print("Check your AWS credentials or permissions and try again.")
            sys.exit(1)

    def mirror_image(self, src_uri, dest_uri):
        debug_print(f"Preparing to mirror image {src_uri} to {dest_uri}...")

        debug_print(f"Pulling image {src_uri}...")
        try:
            self.docker_client.images.pull(src_uri)
        except Exception as e:
            error_print(f"Error pulling image {src_uri}: {e}")
            error_print("Check the image source and try again.")
            error_print("Reach out to Scale for help with authorization issues.")
            sys.exit(1)

        try:
            ecr_registry, ecr_repo, ecr_tag = get_registry_repository_tag(dest_uri)
            ecr_reg_repo = f"{ecr_registry}/{ecr_repo}"
            debug_print(f"Tagging image {src_uri} to repo {ecr_reg_repo} and tag {ecr_tag}...")
            try:
                self.docker_client.images.get(src_uri).tag(ecr_reg_repo, tag=ecr_tag)
            except Exception as e:
                error_print(f"Error tagging image {src_uri} as {dest_uri}: {e}")
                error_print("Check your permissions and try again.")
                sys.exit(1)

            debug_print(f"Pushing image {dest_uri}...")
            debug_print("Depending on image size, this could take some time...")
            try:
                push_log = str(self.docker_client.images.push(dest_uri))
                for line in push_log.split("\n"):
                    if "error" in line.lower():
                        raise Exception(line)
            except Exception as e:
                error_print(f"Error pushing image {dest_uri}: {e}")
                error_print("Check your permissions and try again.")
                sys.exit(1)
        finally:
            try:
                self.docker_client.images.remove(src_uri)
            except Exception as e:
                error_print(f"Error removing locally downloaded image {src_uri}: {e}")
                error_print("Try to remove manually.")

        debug_print("Finished mirroring image.")

    def upload_image_from_tar(self, tar_file_path, dest_uri):
        debug_print(f"Preparing to upload image from tar file {tar_file_path} to {dest_uri}...")

        debug_print(f"Loading image from tar file {tar_file_path}...")
        try_count = 0
        while True:
            try:
                with open(tar_file_path, 'rb') as tar_file:
                    load_response = self.docker_client.images.load(tar_file)
                    image_id = load_response[0].id
                break
            except Exception as e:
                try_count += 1
                if try_count > 3:
                    error_print(f"Error loading image from tar file {tar_file_path}: {e}")
                    error_print("Check the tar file source for errors or expiration.")
                    sys.exit(1)
                debug_print(f"Error loading image from tar file {tar_file_path}: {e}")
                debug_print(f"Retrying loading image from tar file {tar_file_path} {try_count} time(s)...")
                time.sleep(5)
        debug_print(f"Loaded image locally with id {image_id}.")

        debug_print(f"Pushing image to {dest_uri}...")
        debug_print("Depending on image size, this could take some time...")
        try:
            self.docker_client.images.get(image_id).tag(dest_uri)
            push_log = str(self.docker_client.images.push(dest_uri))
            for line in push_log.split("\n"):
                if "error" in line.lower():
                    raise Exception(line)
        except Exception as e:
            error_print(f"Error pushing image {dest_uri}: {e}")
            error_print("Check your permissions and try again.")
            sys.exit(1)
        finally:
            try:
                self.docker_client.images.remove(image_id, force=True)
            except Exception as e:
                error_print(f"Error removing locally loaded image {image_id}: {e}")
                error_print("Try to remove manually.")

        debug_print("Finished uploading image from tar file.")


def download_uri(uri):
    if uri.startswith("http"):
        uri_hash = get_uri_hash(uri)
        out_path = os.path.join(CACHE_DIR, uri_hash)

        if not os.path.exists(out_path):
            os.makedirs(CACHE_DIR, exist_ok=True)

            tmp_out_path = f"{out_path}.tmp"
            try:
                debug_print(f"Downloading {uri} to {out_path}...")
                debug_print("Depending on file size, this could take some time...")
                with requests.get(uri, stream=True) as r:
                    with open(tmp_out_path, 'wb') as f:
                        shutil.copyfileobj(r.raw, f)
                os.rename(tmp_out_path, out_path)
                debug_print(f"Finished downloading to {out_path}.")
            except Exception as e:
                error_print(f"Error downloading {uri} to {out_path}: {e}")
                error_print("Check the source for errors or if using a presigned URL, check for expiration.")
                sys.exit(1)
            finally:
                # Delete the temp file if it exists
                if os.path.exists(tmp_out_path):
                    os.remove(tmp_out_path)
        else:
            debug_print(f"Using cached file for {uri} at {out_path}.")

        return out_path
    else:
        raise ValueError("Currently only supports http(s) URIs")


def download_and_untar_uri(uri):
    uri_hash = get_uri_hash(uri)
    unpack_dir = os.path.join(CACHE_UNTAR_DIR, uri_hash)

    if not os.path.exists(unpack_dir):
        tar_file_path = download_uri(uri)

        temp_unpack_dir = f"{unpack_dir}-tmp"
        os.makedirs(temp_unpack_dir, exist_ok=True)

        debug_print(f"Extracting tar file {tar_file_path} to {unpack_dir}...")
        try:
            with tarfile.open(tar_file_path) as tar:
                tar.extractall(path=temp_unpack_dir)
            os.rename(temp_unpack_dir, unpack_dir)
            debug_print(f"Finished extracting to {unpack_dir}.")
        except Exception as e:
            error_print(f"Error extracting tar file {tar_file_path}")
            error_print(e)
            error_print("Check the tar file source for errors or if using a presigned URL, check for expiration.")
            sys.exit(1)
        finally:
            # Delete the temp directory if it exists
            if os.path.exists(temp_unpack_dir):
                shutil.rmtree(temp_unpack_dir)
    else:
        debug_print(f"Using cached directory for {uri} at {unpack_dir}.")

    return unpack_dir


def get_registry_repository_tag(image_uri):
    registry, repository_and_tag = image_uri.split("/", 1)
    repository, tag = repository_and_tag.split(":")
    return registry, repository, tag


def split_s3_uri(s3_uri):
    parsed = urlparse(s3_uri)
    bucket = parsed.netloc
    key = parsed.path.lstrip('/')
    return bucket, key


def is_s3_presigned_url(uri):
    pattern = r'^https://.+\.s3\..+\.amazonaws\.com/.+$'
    return bool(re.match(pattern, uri))


def remove_uri_query(uri):
    return uri.split("?")[0]


def get_uri_hash(uri):
    if is_s3_presigned_url(uri):
        uri = remove_uri_query(uri)
    return generate_sha256_hash(uri)


def generate_sha256_hash(s):
    return hashlib.sha256(s.encode()).hexdigest()



from typing import Optional
from scale_egp.sdk.types.user_info import UserInfoResponse

from scale_egp.utils.api_utils import APIEngine


class UsersCollection(APIEngine):
    """
    Collections class for Scale EGP users.
    """

    _sub_path = "users"

    def who_am_i(self) -> UserInfoResponse:
        """
        Get the currently authenticated user.

        Returns:
            The currently authenticated user.
        """
        response = self._get(sub_path="user-info")
        return UserInfoResponse.from_dict(response.json())

    def get(self, id: str) -> UserInfoResponse:
        """
                Get a user by ID.

                Args:
                    id: The ID of the user.
        >
                Returns:
                    The user.
        """
        response = self._get(sub_path=f"{self._sub_path}/{id}")
        return UserInfoResponse.from_dict(response.json())

    def get_default_account_id(self) -> Optional[str]:
        user_info = self.who_am_i()
        candidates = [u.account.id for u in user_info.accounts if u.role == "admin"]
        if len(candidates) > 0:
            return candidates[0]
        # No account with admin role found. This shouldn't happen as each
        # user should have a personal account, but if this does happen,
        # we can return the first account in the list
        if len(user_info.accounts) > 0:
            return user_info.accounts[0].account.id
        # No default account could be identified
        return None



import json
from typing import Any, Dict

from pydantic import BaseModel
from scale_egp.cli.formatter import Markdownable
from scale_egp.sdk.enums import ModelVendor
from scale_egp.sdk.types.model_templates import ModelTemplate
from scale_egp.sdk.types.models import (
    ModelInstance,
)


class ModelInstanceDescription(BaseModel, Markdownable):
    model_instance: ModelInstance
    model_template: ModelTemplate

    def _get_description_dict(self) -> Dict[str, Any]:
        return {
            "model_instance": json.loads(self.model_instance.json()),
            "model_template": json.loads(self.model_template.json()),
        }

    def to_markdown(self) -> str:
        return (
            f"# {self.model_instance.name} (id: {self.model_instance.id})\n"
            f"\n"
            f"*type*: {self.model_template.model_type.value}\n"
            f"*vendor*: {(self.model_instance.model_vendor or ModelVendor.LAUNCH).value}\n"
            f"\n"
            f"{self.model_instance.description or ''}\n"
            f"## Model request schema\n"
            f"```\n"
            f"{json.dumps(self.model_instance.request_schema, indent=2)}\n"
            f"```\n"
            f"## Model response schema\n"
            f"```\n"
            f"{json.dumps(self.model_instance.response_schema, indent=2)}\n"
            f"```\n"
        )

    def json(self, **dumps_kwargs: Any) -> str:
        return json.dumps(self._get_description_dict(), **dumps_kwargs)



from typing import List, Optional


from scale_egp.sdk.types.questions import CategoricalChoice, Question, QuestionRequest
from scale_egp.utils.api_utils import (
    APIEngine,
)
from scale_egp.sdk.enums import QuestionType


class QuestionCollection(APIEngine):
    _sub_path = "v2/questions"

    def create(
        self,
        type: QuestionType,
        title: str,
        prompt: str,
        choices: Optional[List[CategoricalChoice]],
        multi: Optional[bool] = None,
        dropdown: Optional[bool] = False,
        required: Optional[bool] = False,
        conditions: Optional[List[dict]] = None,
        account_id: Optional[str] = None,
    ) -> Question:
        """
        Create a new Question.

        Args:
            type: The type of the Question.
            title: The title of the Question.
            prompt: The prompt of the Question.
            account_id: The ID of the account to create this Question for.
            choices: The choices of the Question.
            multi: Whether the question is multi-select
            dropdown: Whether the question is to be displayed as a dropdown
            required: Whether the question is required
            conditions: The conditions for the question
        Returns:
            The newly created Evaluation Configuration.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=QuestionRequest(
                type=type.value,
                title=title,
                prompt=prompt,
                account_id=account_id or self._api_client.account_id,
                choices=[choice.to_dict() for choice in choices] if choices else None,
                multi=multi,
                dropdown=dropdown,
                required=required,
                conditions=conditions,
            ),
        )
        return Question.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> Question:
        """
        Get the details of a Question.

        Args:
            id: The ID of the Question.

        Returns:
            The Question.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return Question.from_dict(response.json())

    def list(
        self,
    ) -> List[Question]:
        """
        List Questions.

        Returns:
            A list of Questions.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [Question.from_dict(evaluation_config) for evaluation_config in response.json()]




from __future__ import annotations

from typing import Optional, Dict, Any, Union, List, Literal

from pydantic import Field, validator, ValidationError

from scale_egp.sdk.enums import TestCaseSchemaType, ExtraInfoSchemaType
from scale_egp.utils.model_utils import Entity, BaseModel, RootModel


class ChunkData(BaseModel):
    text: str
    metadata: Dict[str, Any]

class StringExtraInfo(BaseModel):
    schema_type: Literal[ExtraInfoSchemaType.STRING] = ExtraInfoSchemaType.STRING
    info: str

class ChunksExtraInfo(BaseModel):
    schema_type: Literal[ExtraInfoSchemaType.CHUNKS] = ExtraInfoSchemaType.CHUNKS
    chunks: List[ChunkData]

class ExtraInfo(RootModel):
    __root__: Union[StringExtraInfo, ChunksExtraInfo] = Field(..., discriminator="schema_type")

class GenerationTestCaseData(BaseModel):
    """
    A data model representing the data of a Testcase with a GENERATION schema type.

    Attributes:
        input: The input to the generation model
        expected_output: The expected output of the generation model
        expected_extra_info: The expected extra info of the generation model
    """

    input: str
    expected_output: Optional[str] = ""
    expected_extra_info: Optional[ExtraInfo]


class TestCase(Entity):
    """
    A data model representing a test case.

    Attributes:
        id: The ID of the test case
        evaluation_dataset_id: The ID of the evaluation dataset that the test case belongs to
        schema_type: The schema type of the test case
        test_case_data: The data of the test case. In general, this can be thought of as the
            data to evaluate an application against.
    """

    id: str
    evaluation_dataset_id: str
    schema_type: TestCaseSchemaType
    test_case_data: Union[GenerationTestCaseData]
    test_case_metadata: Optional[Dict[str, Any]]
    chat_history: Optional[Dict[str, Any]]


class TestCaseRequest(BaseModel):
    schema_type: TestCaseSchemaType
    test_case_data: Union[GenerationTestCaseData]
    account_id: Optional[str] = Field(
        description="Account to create knowledge base in. If you have access to more than one "
        "account, you must specify an account_id"
    )
    test_case_metadata: Optional[Dict[str, Any]] = Field(
        description="Additional metadata to associate with the test case"
    )
    chat_history: Optional[Dict[str, Any]] = Field(
        description="Used for tracking previous chat interactions for multi-chat test cases"
    )

    @validator("test_case_data")
    @classmethod
    def test_case_data_matches_schema_type(cls, test_case_data, values):
        schema_type = values.get("schema_type")
        TestCaseSchemaValidator.validate(schema_type, test_case_data)
        return test_case_data


class TestCaseSchemaValidator:
    TEST_CASE_SCHEMA_TO_DATA_TYPE = {
        TestCaseSchemaType.GENERATION: GenerationTestCaseData,
    }

    @classmethod
    def validate(cls, schema_type: TestCaseSchemaType, data: Dict[str, Any]):
        try:
            model = cls.TEST_CASE_SCHEMA_TO_DATA_TYPE[schema_type]
            model.from_dict(data)
        except ValidationError as e:
            raise ValueError(f"Test case data does not match schema type {schema_type}") from e

    @classmethod
    def get_model(cls, schema_type: TestCaseSchemaType):
        return cls.TEST_CASE_SCHEMA_TO_DATA_TYPE[schema_type]

    @classmethod
    def dict_to_model(cls, schema_type: TestCaseSchemaType, data: Dict[str, Any]):
        return cls.TEST_CASE_SCHEMA_TO_DATA_TYPE[schema_type].from_dict(data)



from datetime import datetime, timedelta
import json
from typing import List, Optional, Dict, Any, Union

import httpx

from scale_egp.sdk.enums import ChunkUploadStatus
from scale_egp.sdk.enums import EmbeddingModelName, EmbeddingConfigType
from scale_egp.sdk.types.chunks import (
    Chunk,
)
from scale_egp.sdk.types.embeddings import EmbeddingConfigModelsAPI, EmbeddingConfigBase
from scale_egp.sdk.types.knowledge_base_artifacts import (
    KnowledgeBaseArtifact,
    ListKnowledgeBaseArtifactsResponse,
)
from scale_egp.sdk.types.knowledge_base_chunks import (
    KnowledgeBaseChunksResponse,
    KnowledgeBaseQueryRequest,
    KnowledgeBaseQueryResponse,
)
from scale_egp.sdk.types.knowledge_base_uploads import (
    KnowledgeBaseDataSource,
    KnowledgeBaseDataSourceRequest,
    KnowledgeBaseUpload,
    KnowledgeBaseRemoteUploadRequest,
    KnowledgeBaseLocalChunkUploadRequest,
    KnowledgeBaseUploadResponse,
    KnowledgeBaseUploadSchedule,
    KnowledgeBaseUploadSchedulePauseRequest,
    KnowledgeBaseUploadScheduleRequest,
    ListKnowledgeBaseUploadsResponse,
    CancelKnowledgeBaseUploadResponse,
    LocalChunksSourceConfig,
    ChunkingStrategyConfig,
    ChunkToUpload,
    DataSourceAuthConfig,
    RemoteDataSourceConfig,
)
from scale_egp.sdk.types.knowledge_bases import (
    KnowledgeBase,
    KnowledgeBaseRequest,
    KnowledgeBaseResponse,
    ListKnowledgeBasesResponse,
)
from scale_egp.utils.api_utils import APIEngine


class KnowledgeBaseCollection(APIEngine):
    _sub_path = "v2/knowledge-bases"

    def uploads(self) -> "KnowledgeBaseUploadsCollection":
        """
        Returns a KnowledgeBaseUploadsCollection object for uploads associated with a knowledge
        base.

        Returns:
            A KnowledgeBaseUploadsCollection object.
        """
        return KnowledgeBaseUploadsCollection(self._api_client)

    def artifacts(self) -> "KnowledgeBaseArtifactsCollection":
        """
        Returns a KnowledgeBaseArtifactsCollection object for artifacts associated with a
        knowledge base.

        Returns:
            A KnowledgeBaseArtifactsCollection object.
        """
        return KnowledgeBaseArtifactsCollection(self._api_client)

    def chunks(self) -> "KnowledgeBaseChunksCollection":
        """
        Returns a KnowledgeBaseChunksCollection object for chunks associated with a
        knowledge base.

        Returns:
            A KnowledgeBaseChunksCollection object.
        """
        return KnowledgeBaseChunksCollection(self._api_client)
    
    def upload_schedules(self) -> "KnowledgeBaseUploadScheduleCollection":
        """
        Returns a KnowledgeBaseUploadScheduleCollection object for upload schedules associated with a
        knowledge base.

        Returns:
            A KnowledgeBaseUploadScheduleCollection object.
        """
        return KnowledgeBaseUploadScheduleCollection(self._api_client)

    def create(
        self,
        name: str,
        embedding_model_name: Optional[EmbeddingModelName] = None,
        model_deployment_id: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        account_id: Optional[str] = None,
    ) -> KnowledgeBase:
        """
        Create a new Knowledge Base. Must pass either embedding_model_name or model_deployment_id.

        Args:
            name: The name of the Knowledge Base.
            embedding_model_name: The name of the embedding model to use for the Knowledge Base.
            model_deployment_id: ID for a EmbeddingConfigModelsAPI config.
            metadata: The metadata of the Knowledge Base.
            account_id: The ID of the account to create this Knowledge Base for.

        Returns:
            The newly created Knowledge Base.
        """

        if embedding_model_name is not None and model_deployment_id is not None:
            raise ValueError(
                "Must pass either embedding_model_name or model_deployment_id, not both."
            )
        elif embedding_model_name is not None:
            embedding_config = EmbeddingConfigBase(
                type=EmbeddingConfigType.BASE, embedding_model=embedding_model_name
            )
        elif model_deployment_id is not None:
            embedding_config = EmbeddingConfigModelsAPI(
                type=EmbeddingConfigType.MODELS_API, model_deployment_id=model_deployment_id
            )
        else:
            raise ValueError("Must pass either embedding_model_name or model_deployment_id.")

        response = self._post(
            sub_path=self._sub_path,
            request=KnowledgeBaseRequest(
                account_id=account_id or self._api_client.account_id,
                knowledge_base_name=name,
                embedding_config=embedding_config,
                metadata=metadata,
            ),
        )
        response_model = KnowledgeBaseResponse.from_dict(response.json())
        return self.get(id=response_model.knowledge_base_id)

    def get(
        self,
        id: str,
    ) -> KnowledgeBase:
        """
        Get an Knowledge Base by ID.

        Args:
            id: The ID of the Knowledge Base.

        Returns:
            The Knowledge Base.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return KnowledgeBase.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Delete a Knowledge Base by ID.

        Args:
            id: The ID of the Knowledge Base.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[KnowledgeBase]:
        """
        List all Knowledge Bases.

        Returns:
            A list of Knowledge Bases.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        response_model = ListKnowledgeBasesResponse.from_dict(response.json())
        return response_model.items


class KnowledgeBaseArtifactsCollection(APIEngine):
    _sub_path = "v2/knowledge-bases/{knowledge_base_id}/artifacts"

    def get(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
        status_filter: Optional[ChunkUploadStatus] = ChunkUploadStatus.COMPLETED.value,
    ) -> KnowledgeBaseArtifact:
        """
        Get a Knowledge Base Artifact by ID.

        Args:
            id: The ID of the Knowledge Base Artifact.
            knowledge_base: The Knowledge Base the artifact was created for.
            status_filter: Return only artifacts with the given status.

        Returns:
            The Knowledge Base Artifact.
        """
        response = self._get(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/"
            f"{id}",
            query_params=dict(
                status_filter=status_filter,
            ),
        )
        return KnowledgeBaseArtifact.from_dict(response.json())

    def list(
        self,
        knowledge_base: KnowledgeBase,
    ) -> List[KnowledgeBaseArtifact]:
        """
        List all Knowledge Base Artifacts.

        Returns:
            A list of Knowledge Base Artifacts.
        """
        response = self._get(
            sub_path=self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id),
        )
        response_model = ListKnowledgeBaseArtifactsResponse.from_dict(response.json())
        return response_model.artifacts


class KnowledgeBaseUploadsCollection(APIEngine):
    _sub_path = "v2/knowledge-bases/{knowledge_base_id}/uploads"

    def create_remote_upload(
        self,
        knowledge_base: KnowledgeBase,
        data_source_config: RemoteDataSourceConfig,
        data_source_auth_config: Optional[DataSourceAuthConfig],
        chunking_strategy_config: ChunkingStrategyConfig,
    ) -> KnowledgeBaseUpload:
        """
        Create a new remote upload.

        Args:
            knowledge_base: The Knowledge Base to upload data to.
            data_source_config: The data source config.
            data_source_auth_config: The data source auth config.
            chunking_strategy_config: The chunking strategy config.

        Returns:
            The newly created remote upload.
        """
        response = self._post(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}",
            request=KnowledgeBaseRemoteUploadRequest(
                data_source_config=data_source_config,
                data_source_auth_config=data_source_auth_config,
                chunking_strategy_config=chunking_strategy_config,
            ),
        )
        response_model = KnowledgeBaseUploadResponse.from_dict(response.json())
        return self.get(id=response_model.upload_id, knowledge_base=knowledge_base)

    def create_local_upload(
        self,
        knowledge_base: KnowledgeBase,
        data_source_config: LocalChunksSourceConfig,
        chunks: List[ChunkToUpload],
    ) -> KnowledgeBaseUpload:
        """
        Create a new local upload.

        Args:
            knowledge_base: The Knowledge Base to upload data to.
            data_source_config: The data source config.
            chunks: The chunks to upload.

        Returns:
            The newly created local upload.
        """
        response = self._post(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}",
            request=KnowledgeBaseLocalChunkUploadRequest(
                data_source_config=data_source_config,
                chunks=chunks,
            ),
        )
        response_model = KnowledgeBaseUploadResponse.from_dict(response.json())
        return self.get(id=response_model.upload_id, knowledge_base=knowledge_base)

    def get(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
    ) -> KnowledgeBaseUpload:
        """
        Get an Knowledge Base Upload by ID.

        Args:
            id: The ID of the Knowledge Base Upload.
            knowledge_base: The Knowledge Base the upload was created for.

        Returns:
            The Knowledge Base Upload.
        """
        response = self._get(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/"
            f"{id}",
        )
        return KnowledgeBaseUpload.from_dict(response.json())

    def list(
        self,
        knowledge_base: KnowledgeBase,
    ) -> List[KnowledgeBaseUpload]:
        """
        List all Knowledge Base Uploads.

        Returns:
            A list of Knowledge Base Uploads.
        """
        response = self._get(
            sub_path=self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id),
        )

        # TODO: This is a hack. Server side field names should be consolidated.
        json_response = response.json()
        for i in range(len(json_response.get("uploads", []))):
            json_response["uploads"][i]["upload_id"] = json_response["uploads"][i]["id"]
        response_model = ListKnowledgeBaseUploadsResponse.from_dict(json_response)
        return response_model.uploads

    def cancel(
        self,
        knowledge_base: KnowledgeBase,
        id: str,
    ) -> bool:
        """
        Cancel an upload.

        Args:
            knowledge_base: The Knowledge Base the upload was created for.
            id: The ID of the upload to cancel.

        Returns:
            True if the upload was canceled, False otherwise.
        """
        response = self._post(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/"
            f"{id}/cancel",
            request=None,
        )
        response_model = CancelKnowledgeBaseUploadResponse.from_dict(response.json())
        return response_model.canceled


class KnowledgeBaseChunksCollection(APIEngine):
    _sub_path = "v2/knowledge-bases/{knowledge_base_id}"

    def query(
        self,
        knowledge_base: KnowledgeBase,
        query: str,
        top_k: int,
        include_embeddings: bool = False,
        metadata_filters: Optional[Dict[str, Any]] = None,
        use_exact_knn: bool = False,
    ) -> List[Chunk]:
        """
        Query a Knowledge Base.

        Args:
            knowledge_base: The Knowledge Base to query.
            query: The query string.
            top_k: The number of results to return.
            include_embeddings: Whether to include embeddings in the response.
            metadata_filters: The metadata to filter query results by. This approach uses a Faiss
                engine with an HNSW algorithm filtering during the k-NN search, as opposed to
                before or after the k-NN search, which ensures that k results are returned (if
                there are at least k results in total).
            use_exact_knn: Whether or not to use exact_knn search (only with top_k <= 1000)

        Returns:
            The query response.
        """

        if use_exact_knn and top_k > 1000:
            raise ValueError("The maximum number of chunks allowed for exact_knn is 1000.")

        path = f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/query"
        if use_exact_knn:
            path = f"{path}?use_exact_knn=true"

        response = self._post(
            sub_path=path,
            request=KnowledgeBaseQueryRequest(
                query=query,
                top_k=top_k,
                include_embeddings=include_embeddings,
                metadata_filters=metadata_filters,
            ),
            # Bump timeout on the knowledge_bases request:
            #   Reasoning for 360 - we set a default timeout of 300s on the opensearch API request,
            #   so we should set this higher than that, otherwise we risk a scenario in which opensearch succeeds
            #   but our API fails.
            timeout=360,
        )
        response_model = KnowledgeBaseQueryResponse.from_dict(response.json())
        return response_model.chunks

    def get(
        self,
        knowledge_base: KnowledgeBase,
        chunk_id: Optional[str] = None,
        metadata_filters: Optional[Dict[str, Any]] = None,
    ) -> List[Chunk]:
        """
        Get chunks from a Knowledge Base.

        Args:
            knowledge_base: The Knowledge Base to query.
            chunk_id: The chunk ID to match.
            metadata_filters: The metadata whose values to match.

        Returns:
            A list of Chunks.
        """
        query_params = dict()
        if chunk_id:
            query_params["chunk_id"] = chunk_id
        if metadata_filters:
            query_params["metadata_filters"] = json.dumps(metadata_filters)
        response = self._get(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/chunks",
            query_params=query_params,
        )
        response_model = KnowledgeBaseChunksResponse.from_dict(response.json())
        return response_model.chunks


class KnowledgeBaseDataSourceCollection(APIEngine):
    _sub_path = "v3/knowledge-base-data-sources"

    def create(
        self,
        name: str,
        data_source_config: RemoteDataSourceConfig,
        data_source_auth_config: Optional[DataSourceAuthConfig] = None,
        description: Optional[str] = None,
        account_id: Optional[str] = None,
    ) -> KnowledgeBaseDataSource:
        """
        Create a new Knowledge Base Data Source.

        Args:
            name: The name of the data source.
            data_source_config: The data source config.
            data_source_auth_config: The data source auth config.
            description: The description of the data source.
            account_id: The ID of the account to create this data source for.

        Returns:
            The newly created data source.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=KnowledgeBaseDataSourceRequest(
                name=name,
                data_source_config=data_source_config,
                data_source_auth_config=data_source_auth_config,
                description=description,
                account_id=account_id or self._api_client.account_id,
            ),
        )
        return KnowledgeBaseDataSource.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> KnowledgeBaseDataSource:
        """
        Get a Knowledge Base Data Source by ID.

        Args:
            id: The ID of the Knowledge Base Data Source.

        Returns:
            The Knowledge Base Data Source.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return KnowledgeBaseDataSource.from_dict(response.json())

    def list(
        self,
        account_id: Optional[str] = None,
    ) -> List[KnowledgeBaseDataSource]:
        """
        List all Knowledge Base Data Sources for the given account.

        Args:
            account_id: The ID of the account to list data sources for. Defaults to the account
                associated with the API client.

        Returns:
            A list of Knowledge Base Data Sources.
        """
        response = self._get(
            sub_path=self._sub_path,
            query_params=dict(account_id=account_id or self._api_client.account_id),
        )

        return [KnowledgeBaseDataSource.from_dict(data_source) for data_source in response.json()]

    def verify(
        self,
        knowledge_base_data_source: KnowledgeBaseDataSource,
    ) -> bool:
        """
        Verifies the data source's auth configuration.

        Args:
            knowledge_base_data_source: The Knowledge Base Data Source entity.

        Returns:
            True if the data source auth config is valid, False otherwise.
        """
        response = self._post(
            sub_path=f"{self._sub_path}/{knowledge_base_data_source.id}/verify",
        )
        return response.status_code == httpx.codes.ok

    def update(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        data_source_auth_config: Optional[DataSourceAuthConfig] = None,
    ) -> KnowledgeBaseDataSource:
        """
        Update a Knowledge Base Data Source.

        Args:
            id: The ID of the Knowledge Base Data Source.
            name: The updated name of the data source.
            description: The updated description of the data source.
            data_source_auth_config: The updated data source auth config.

        Returns:
            The updated Knowledge Base Data Source.
        """
        response = self._patch(
            sub_path=f"{self._sub_path}/{id}",
            request=KnowledgeBaseDataSourceRequest.partial(
                name=name,
                description=description,
                data_source_auth_config=data_source_auth_config,
            ),
        )
        return KnowledgeBaseDataSource.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Delete a Knowledge Base Data Source.

        Args:
            id: The ID of the Knowledge Base Data Source.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok


class KnowledgeBaseUploadScheduleCollection(APIEngine):
    _sub_path = "v3/knowledge-bases/{knowledge_base_id}/upload-schedules"

    def create(
        self,
        knowledge_base: KnowledgeBase,
        knowledge_base_data_source: KnowledgeBaseDataSource,
        chunking_strategy_config: ChunkingStrategyConfig,
        interval: timedelta,
        next_run_at: Optional[datetime] = None,
    ) -> KnowledgeBaseUploadSchedule:
        """
        Create a new Knowledge Base Upload Schedule.

        Args:
            knowledge_base: The Knowledge Base to upload data to.
            knowledge_base_data_source: The Knowledge Base Data Source to use for the upload.
            chunking_strategy_config: The chunking strategy config.
            interval: The interval between uploads.
            next_run_at: The time of the next upload.

        Returns:
            The newly created upload schedule.
        """
        response = self._post(
            sub_path=self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id),
            request=KnowledgeBaseUploadScheduleRequest(
                knowledge_base_data_source_id=knowledge_base_data_source.id,
                chunking_strategy_config=chunking_strategy_config,
                interval=interval.total_seconds(),
                next_run_at=next_run_at.isoformat() if next_run_at else None,
                account_id=knowledge_base_data_source.account_id,
            ),
        )
        return KnowledgeBaseUploadSchedule.from_dict(response.json())

    def get(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
    ) -> KnowledgeBaseUploadSchedule:
        """
        Get an Knowledge Base Upload Schedule by ID.

        Args:
            id: The ID of the Knowledge Base Upload Schedule.
            knowledge_base: The Knowledge Base the upload schedule was created for.

        Returns:
            The Knowledge Base Upload Schedule.
        """
        response = self._get(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/{id}"
        )
        return KnowledgeBaseUploadSchedule.from_dict(response.json())

    def list(
        self,
        knowledge_base: KnowledgeBase,
    ) -> List[KnowledgeBaseUploadSchedule]:
        """
        List all Knowledge Base Upload Schedules for a given Knowledge Base.

        Args:
            knowledge_base: The Knowledge Base to list upload schedules for.

        Returns:
            A list of Knowledge Base Upload Schedules.
        """
        response = self._get(
            sub_path=self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id),
        )

        return [
            KnowledgeBaseUploadSchedule.from_dict(upload_schedule)
            for upload_schedule in response.json()
        ]

    def update(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
        chunking_strategy_config: Optional[ChunkingStrategyConfig] = None,
        interval: Optional[timedelta] = None,
        next_run_at: Optional[datetime] = None,
    ) -> KnowledgeBaseUploadSchedule:
        """
        Update a Knowledge Base Upload Schedule.

        Args:
            id: The ID of the Knowledge Base Upload Schedule.
            knowledge_base: The Knowledge Base the upload schedule was created for.
            chunking_strategy_config: The updated chunking strategy config.
            interval: The updated interval between uploads.
            next_run_at: The updated time of the next upload (will resume the scheduled uploads if paused).

        Returns:
            The updated Knowledge Base Upload Schedule.
        """
        response = self._patch(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/{id}",
            request=KnowledgeBaseUploadScheduleRequest.partial(
                chunking_strategy_config=chunking_strategy_config,
                interval=interval.total_seconds() if interval else None,
                next_run_at=next_run_at.isoformat() if next_run_at else None,
            ),
        )
        return KnowledgeBaseUploadSchedule.from_dict(response.json())

    def pause(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
    ) -> KnowledgeBaseUploadSchedule:
        """
        Pause a Knowledge Base Upload Schedule.

        Args:
            id: The ID of the Knowledge Base Upload Schedule.
            knowledge_base: The Knowledge Base the upload schedule was created for.

        Returns:
            The updated Knowledge Base Upload Schedule.
        """
        response = self._patch(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/{id}",
            request=KnowledgeBaseUploadSchedulePauseRequest(),
        )
        return KnowledgeBaseUploadSchedule.from_dict(response.json())

    def resume(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
        next_run_at: datetime,
    ) -> KnowledgeBaseUploadSchedule:
        """
        Pause a Knowledge Base Upload Schedule.

        Args:
            id: The ID of the Knowledge Base Upload Schedule.
            knowledge_base: The Knowledge Base the upload schedule was created for.
            next_run_at: The time to run the next upload.

        Returns:
            The updated Knowledge Base Upload Schedule.
        """
        response = self._patch(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/{id}",
            request=KnowledgeBaseUploadScheduleRequest.partial(
                next_run_at=next_run_at.isoformat(),
            ),
        )
        return KnowledgeBaseUploadSchedule.from_dict(response.json())

    def delete(
        self,
        id: str,
        knowledge_base: KnowledgeBase,
    ) -> bool:
        """
        Delete a Knowledge Base Upload Schedule.

        Args:
            id: The ID of the Knowledge Base Upload Schedule.
            knowledge_base: The Knowledge Base the upload schedule was created for.
        """
        response = self._delete(
            sub_path=f"{self._sub_path.format(knowledge_base_id=knowledge_base.knowledge_base_id)}/{id}",
        )
        return response.status_code == httpx.codes.ok



from typing import Any, Optional, List, Dict

import httpx

from scale_egp.sdk.types.evaluation_configs import EvaluationConfig, EvaluationConfigRequest
from scale_egp.sdk.types.question_sets import QuestionSet
from scale_egp.utils.api_utils import (
    APIEngine,
)
from scale_egp.sdk.enums import EvaluationType


class EvaluationConfigCollection(APIEngine):
    _sub_path = "v2/evaluation-configs"

    def create(
        self,
        evaluation_type: EvaluationType,
        question_set: QuestionSet,
        account_id: Optional[str] = None,
    ) -> EvaluationConfig:
        """
        Create a new Evaluation Configuration.

        Args:
            evaluation_type: The type of Evaluation. Only `HUMAN` is supported.
            question_set: The Question Set to associate with the Evaluation.
            account_id: The ID of the account to create this Evaluation Configuration for.

        Returns:
            The newly created Evaluation Configuration.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=EvaluationConfigRequest(
                evaluation_type=evaluation_type.value,
                question_set_id=question_set.id,
                account_id=account_id or self._api_client.account_id,
            ),
        )
        return EvaluationConfig.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> EvaluationConfig:
        """
        Get the details of an evaluation config.

        Args:
            id: The ID of the Evaluation Configuration.

        Returns:
            The Evaluation Configuration.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return EvaluationConfig.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Delete an Evaluation Configuration.

        Args:
            id: The ID of the Evaluation Configuration.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[EvaluationConfig]:
        """
        List Evaluation Configurations.

        Returns:
            A list of Evaluation Configurations.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [EvaluationConfig.from_dict(evaluation_config) for evaluation_config in response.json()]




from __future__ import annotations

from datetime import datetime
from typing import Optional, Dict, Any

from pydantic import Field

from scale_egp.sdk.enums import EvaluationStatus
from scale_egp.utils.model_utils import Entity, BaseModel


class Evaluation(Entity):
    """
    A data model representing an evaluation.

    Attributes:
        id: The ID of the evaluation
        name: The name of the evaluation
        description: The description of the evaluation
        status: The status of the evaluation
        application_spec_id: The ID of the application spec that the evaluation is for
        evaluation_config_id: The ID of the evaluation config
        tags: The tags of the evaluation represented as a dictionary of key value pairs
        created_at: The time the evaluation was created
    """

    id: str
    name: str
    description: str
    status: EvaluationStatus
    application_spec_id: str
    evaluation_config_id: str
    tags: Optional[Dict[str, Any]] = None
    created_at: datetime


class EvaluationRequest(BaseModel):
    name: str
    description: str
    application_spec_id: str
    evaluation_config_id: str
    tags: Optional[Dict[str, Any]] = None
    account_id: Optional[str] = Field(
        description="Account to create knowledge base in. If you have access to more than one "
                    "account, you must specify an account_id"
    )



import os
from abc import ABC, abstractmethod
from typing import Dict, Optional

import httpx

from scale_egp.sdk.collections.application_specs import ApplicationSpecCollection
from scale_egp.sdk.collections.chunks import ChunkCollection
from scale_egp.sdk.collections.completions import CompletionCollection
from scale_egp.sdk.collections.evaluation_configs import EvaluationConfigCollection
from scale_egp.sdk.collections.evaluation_datasets import EvaluationDatasetCollection
from scale_egp.sdk.collections.evaluations import EvaluationCollection
from scale_egp.sdk.collections.knowledge_bases import (
    KnowledgeBaseCollection,
    KnowledgeBaseDataSourceCollection,
)
from scale_egp.sdk.collections.model_groups import ModelGroupCollection
from scale_egp.sdk.collections.model_templates import ModelTemplateCollection
from scale_egp.sdk.collections.models import ModelInstanceCollection
from scale_egp.sdk.collections.question_sets import QuestionSetCollection
from scale_egp.sdk.collections.questions import QuestionCollection
from scale_egp.sdk.collections.users import UsersCollection
from scale_egp.sdk.types.user_info import UserInfoResponse
from scale_egp.utils.api_utils import APIEngine
from scale_egp.utils.model_utils import BaseModel

DEFAULT_ENDPOINT_URL = "https://api.egp.scale.com"


class EGPClientConfig(BaseModel):
    proxies: Optional[Dict[str, httpx.Proxy]] = None
    timeout: Optional[float] = None


class EGPClientConfigGenerator(ABC):
    @abstractmethod
    def generate(self, *args, **kwargs) -> EGPClientConfig:
        pass


class EGPClient:
    """
    The SGP client object. This is the main entry point for interacting with SGP.

    From this client you can access "collections" which interact with various SGP components.
    Each collection will have various methods that interact with the API. Some collections may
    have sub-collections to signify a hierarchical relationship between the entities they represent.

    For users within strict firewall environments, the client can be configured to use a proxy via the config_generator
    argument. Here is an example of how to do Kerberos Authentication through a proxy.

    ```python
    import httpx
    from requests_kerberos import HTTPKerberosAuth

    from scale_egp.sdk.client import EGPClient, EGPClientConfig, EGPClientConfigGenerator

    class KerberosProxyConfigGenerator(EGPClientConfigGenerator):

        def __init__(self, proxy_url: str):
            self._proxy_url = proxy_url

        def generate(self) -> EGPClientConfig:
            return EGPClientConfig(proxies={
                "http://": httpx.Proxy(url=self._proxy_url, headers=self._get_proxy_headers()),
                "https://": httpx.Proxy(url=self._proxy_url, headers=self._get_proxy_headers()),
            })

        def _get_proxy_headers(self) -> httpx.Headers:
            auth = HTTPKerberosAuth()
            negotiate_details = auth.generate_request_header(None, parse_url(self._proxy_url).host, is_preemptive=True)
            return httpx.Headers({"Proxy-Authorization": negotiate_details}, encoding="utf-8")

    client = EGPClient(
        api_key="<API_KEY>",
        config_generator=KerberosProxyConfigGenerator("http://proxy.example.com:3128")
    )
    ```
    """

    def __init__(
        self,
        api_key: str = None,
        account_id: str = None,
        endpoint_url: str = None,
        config_generator: Optional[EGPClientConfigGenerator] = None,
        log_curl_commands: Optional[bool] = None,
    ):
        """
        Args:
            api_key: The SGP API key to use. If not provided, the `EGP_API_KEY` environment
                variable will be used. Enterprise customers of SGP should use the API key provided
                to them by their Scale account manager.
            account_id: The SGP account ID to use. If not provided, the `ACCOUNT_ID` environment
                variable will be used.
            endpoint_url: The SGP endpoint URL to use. If not provided, the `EGP_ENDPOINT_URL`
                environment variable will be used. If that is not set, the default SGP endpoint
                URL `https://api.egp.scale.com` will be used. Enterprise customers of SGP should
                use the endpoint URL provided by their Scale account manager.
            config_generator: An instance of EGPClientConfigGenerator, which must implement a generate function that
                returns an EGPClientConfig object. The client config will be used to inject httpx.Client arguments on
                demand per request. This is useful for dynamically setting proxies, timeouts, etc.
        """
        api_key = api_key or os.environ.get("EGP_API_KEY")
        endpoint_url = endpoint_url or os.environ.get("EGP_ENDPOINT_URL", DEFAULT_ENDPOINT_URL)
        self.log_curl_commands = (
            log_curl_commands
            if log_curl_commands is not None
            else os.environ.get("EGP_LOG_CURL_COMMAND", "").upper() == "TRUE"
        )
        self.api_key = api_key
        self.endpoint_url = endpoint_url if endpoint_url.endswith("/") else endpoint_url + "/"
        self.config_generator = config_generator

        if not self.api_key:
            raise ValueError("No API key provided. Please provide an API key.")
        if not self.endpoint_url:
            raise ValueError("No endpoint URL provided. Please provide an endpoint URL.")

        self.account_id = account_id
        if self.account_id is None:
            self.account_id = os.environ.get("EGP_ACCOUNT_ID")
        if self.account_id is None:
            # TODO: if there are more accounts, taking the first one might not be the most
            #  intuitive logic
            self.account_id = self.users().get_default_account_id()
        if self.account_id is None:
            raise ValueError(
                "Failed to determine default account_id, please provide an account_id parameter."
            )

    def user_info(self) -> UserInfoResponse:
        # TODO: this is an exception, should use collections instead
        print(
            "This method will be deprecated soon. Please use the client.users().who_am_i() method instead."
        )
        api_engine = APIEngine(self)
        response = api_engine._get(sub_path="user-info")  # noqa
        return UserInfoResponse.from_dict(response.json())

    def users(self) -> UsersCollection:
        """
        Returns the Users Collection.

        Use this collection to get information about the currently authenticated user or to get
        information about other users.

        Returns:
            The Users Collection.
        """
        return UsersCollection(self)

    def knowledge_bases(self) -> KnowledgeBaseCollection:
        """
        Returns the Knowledge Base Collection.

        Use this collection to create and manage Knowledge Bases.

        Returns:
            The Knowledge Base Collection.
        """
        return KnowledgeBaseCollection(self)

    def knowledge_base_data_sources(self) -> KnowledgeBaseDataSourceCollection:
        """
        Returns the Knowledge Base Data Source Collection.

        Use this collection to create and manage Knowledge Bases.

        Returns:
            The Knowledge Base Data Source Collection.
        """
        return KnowledgeBaseDataSourceCollection(self)

    def chunks(self) -> ChunkCollection:
        """
        Returns the Chunk Collection.

        Use this collection to create and manage Chunks.

        Returns:
            The Chunk Collection.
        """
        return ChunkCollection(self)

    def completions(self) -> CompletionCollection:
        """
        Returns the Completion Collection.

        Use this collection if you want to make request to an LLM to generate a completion.

        Returns:
            The Completion Collection.
        """
        return CompletionCollection(self)

    def model_templates(self) -> ModelTemplateCollection:
        """
        Returns the Model Template Collection.

        Use this collection to create and manage Model Templates.

        In order to prevent any user from creating any arbitrary model, users with more advanced
        permissions can create Model Templates. Models can only be created from Model Templates.
        This allows power users to create a set of approved models that other users can derive
        from.

        When the model is instantiated from a model template, the settings from the template
        are referenced to reserve the required computing resources, pull the correct docker image,
        etc.

        Returns:
            The Model Template Collection.
        """
        return ModelTemplateCollection(self)

    def models(self) -> ModelInstanceCollection:
        """
        Returns the Model Collection.

        Use this collection to create and manage Models.

        in generative AI applications, there are many types of models that are useful. For
        example, embedding models are useful for translating natural language
        into query-able vector representations, reranking models are useful when a vector
        database's query results need to be re-ranked based on some other criteria, and LLMs
        are useful for generating text from a prompt.

        This collection allows you to create, deploy, and manage any custom model you choose if
        none of the built-in models fit your use case.

        Returns:
            The Model Collection.
        """
        return ModelInstanceCollection(self)

    def model_groups(self) -> ModelGroupCollection:
        """
        Returns the Model Group Collection.

        Use this collection to create and manage Model Groups.

        TODO: Write extensive documentation on Model Groups

        Returns:
            The Model Group Collection.
        """
        return ModelGroupCollection(self)

    def evaluation_datasets(self) -> EvaluationDatasetCollection:
        """
        Returns the Evaluation Dataset Collection.

        Use this collection to create and manage Evaluation Datasets or Test Cases within them.

        Returns:
            The Evaluation Dataset Collection.
        """
        return EvaluationDatasetCollection(self)

    def application_specs(self) -> ApplicationSpecCollection:
        """
        Returns the Application Spec Collection.

        Use this collection to create and manage Application Specs. These are specifications for
        the AI application you are building. They contain information about the AI application
        such as its name and description. They are useful to associate your Evaluations with so
        evaluations can be grouped by application.

        Returns:
            The Application Spec Collection.
        """
        return ApplicationSpecCollection(self)

    def questions(self) -> QuestionCollection:
        """
        Returns the Question Collection.

        Use this collection to create and manage Questions.

        Returns:
            The Question Collection.
        """
        return QuestionCollection(self)

    def question_sets(self) -> QuestionSetCollection:
        """
        Returns the Question Set Collection.

        Use this collection to create and manage Question Sets.

        Returns:
            The Question Set Collection.
        """
        return QuestionSetCollection(self)

    def evaluation_configs(self) -> EvaluationConfigCollection:
        """
        Returns the Evaluation Config Collection.

        Use this collection to manage Evaluation Configurations. Evaluation Configurations
        are used to define the parameters of an evaluation.

        Returns:
            The Evaluation Config Collection.
        """
        return EvaluationConfigCollection(self)

    def evaluations(self) -> EvaluationCollection:
        """
        Returns the Evaluation Collection.

        Use this collection to create and manage Evaluations and Test Case Results.

        Evaluations are used to evaluate the performance of AI applications. Users are
        expected to follow the following procedure to perform an evaluation:

        1. Select an Evaluation Dataset
        2. Iterate through the dataset's Test Cases:
          - For each of these test cases, the user use their AI application to generate output
          data on each test case input prompt.
        3. The user then submits this data as as batch of Test Case Results associated
        with an Evaluation.
        4. Annotators will asynchronously log into the SGL annotation platform to annotate the
        submitted Test Case Results. The annotations will be used to evaluate the performance of
        the AI application.
        5. The submitting user will check back on their Test Case Results to see if the `result`
        field was populated. If so, the evaluation is complete and the user can use the annotation
        data to evaluate the performance of their AI application.

        Returns:
            The Evaluation Collection.
        """
        return EvaluationCollection(self)



# This file is shared between the following (using symlinks):
# - egp-api-backend
# - egp-py (SDK, CLI)
# - model code (dockerized models under packages/egp-api-backend/model_images)
# This generally applies to files under packages/egp-api-backend/egp_api_backend/server/utils/model_schemas
# we do this because they're mostly pydantic models describing the contract between
# SDK / CLI <------ API backend ------> Launch models running in docker containers
# Since these files are using in so many places, you have to be careful what you import:
# - Use relative imports for other files in this directory, since you can't assume to know the full model name.
# - If importing backend / sdk specific classes, wrap the import in a try: ... except ImportError and defined a fallback
#   implementation to use for the imported symbol in other contexts.

from enum import Enum
from typing import Any, Dict, Literal, Optional, Union

from pydantic import Field

try:
    from egp_api_backend.server.api.utils.model_utils import BaseModel, RootModel
except ImportError:
    from pydantic import BaseModel

    RootModel = BaseModel

from .model_enums import ModelVendor


class LaunchFineTuningJobConfiguration(BaseModel):
    vendor: Literal[ModelVendor.LAUNCH] = Field(ModelVendor.LAUNCH)
    hyperparameters: Optional[Dict[str, Any]]
    wandb_config: Optional[Dict[str, Any]]
    suffix: Optional[str]
    output: Optional[str]


class LLMEngineFineTuningJobConfiguration(BaseModel):
    vendor: Literal[ModelVendor.LLMENGINE] = Field(ModelVendor.LLMENGINE)
    hyperparameters: Optional[Dict[str, Any]]
    wandb_config: Optional[Dict[str, Any]]
    suffix: Optional[str]
    output: Optional[str]


class OpenAIFineTuningJobConfiguration(BaseModel):
    vendor: Literal[ModelVendor.OPENAI] = Field(ModelVendor.OPENAI)
    hyperparameters: Optional[Dict[str, Any]]
    suffix: Optional[str]


class FineTuningJobVendorConfiguration(RootModel):
    __root__: Union[
        LaunchFineTuningJobConfiguration,
        LLMEngineFineTuningJobConfiguration,
        OpenAIFineTuningJobConfiguration,
    ] = Field(..., discriminator="vendor")


class FineTuningJobEvent(BaseModel):
    timestamp: Optional[float]
    message: str
    level: str


class FineTuningJobStatus(str, Enum):
    PENDING = "PENDING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    RUNNING = "RUNNING"
    CANCELED = "CANCELED"


class RerankingDatasetItem(BaseModel):
    question: str = Field(...)
    chunk: str = Field(...)
    # True if positive example, False otherwise
    positive: bool = Field(...)


class TrainingDatasetORMSchemaTypeEnum(str, Enum):
    GENERATION = "GENERATION"
    RERANKING_QUESTIONS = "RERANKING_QUESTIONS"



from typing import List, Optional
import httpx

from scale_egp.sdk.types.models_group import ModelGroup, ModelGroupRequest
from scale_egp.utils.api_utils import APIEngine
from scale_egp.utils.model_utils import dict_without_none_values, make_partial_model


PartialModelGroupRequest = make_partial_model(ModelGroupRequest)


class ModelGroupCollection(APIEngine):
    """
    Collections class for SGP Model Groups.
    """

    _sub_path = "v3/model-groups"

    def create(
        self,
        name: str,
        description: Optional[str] = None,
        account_id: Optional[str] = None,
    ) -> str:
        """
        Create a new SGP Model Group.

        Returns:
            The created Model Group ID.
        """
        response = self._post(
            sub_path=self._sub_path,
            request=ModelGroupRequest(
                name=name,
                description=description,
                account_id=account_id or self._api_client.account_id,
            ),
        )
        return ModelGroup.from_dict(response.json())

    def get(
        self,
        id: str,
    ) -> ModelGroup:
        """
        Get a Model Group by ID.

        Returns:
            The Model Group.
        """
        response = self._get(
            sub_path=f"{self._sub_path}/{id}",
        )
        return ModelGroup.from_dict(response.json())

    def update(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> ModelGroup:
        """
        Update a Model Group by ID.

        Returns:
            The updated Model Group.
        """
        response = self._patch(
            sub_path=f"{self._sub_path}/{id}",
            request=PartialModelGroupRequest(
                **dict_without_none_values(
                    dict(
                        name=name,
                        description=description,
                    ),
                )
            ),
        )
        return ModelGroup.from_dict(response.json())

    def delete(
        self,
        id: str,
    ) -> bool:
        """
        Delete a Model Group by ID.

        Returns:
            True if the Model Group was successfully deleted.
        """
        response = self._delete(
            sub_path=f"{self._sub_path}/{id}",
        )
        return response.status_code == httpx.codes.ok

    def list(
        self,
    ) -> List[ModelGroup]:
        """
        List all Model Groups.

        Returns:
            A list of Model Groups.
        """
        response = self._get(
            sub_path=self._sub_path,
        )
        return [ModelGroup.from_dict(model) for model in response.json()]
